import glob
import os
import shutil

import mmcv
import numpy as np
import pytest

from xrprimer.calibration.builder import build_calibrator
from xrprimer.data_structure.camera import FisheyeCameraParameter

input_dir = 'test/data/calibration/test_sview_intrinsic_calibrator'
output_dir = 'test/data/output/calibration/test_sview_intrinsic_calibrator'
eps = 1e-4


@pytest.fixture(scope='module', autouse=True)
def fixture():
    if os.path.exists(output_dir):
        shutil.rmtree(output_dir)
    os.makedirs(output_dir, exist_ok=False)


def test_sview_fisheye_distortion_calibrator():
    fisheye_param = FisheyeCameraParameter.fromfile(
        os.path.join(input_dir, 'init_fisheye_param.json'))
    frame_list = sorted(glob.glob(os.path.join(input_dir, 'images', '*.jpg')))
    # test dict config
    calibrator_config = dict(
        mmcv.Config.fromfile('config/calibration/' +
                             'sview_fisheye_distortion_calibrator.py'))
    calibrator = build_calibrator(calibrator_config)
    calibrated_fisheye_param = calibrator.calibrate(frame_list, fisheye_param)
    assert abs(calibrated_fisheye_param.k1 - 0.0) > eps
    assert abs(calibrated_fisheye_param.k2 - 0.0) > eps
    assert abs(calibrated_fisheye_param.k3 - 0.0) > eps
    assert abs(calibrated_fisheye_param.p1 - 0.0) > eps
    assert abs(calibrated_fisheye_param.p2 - 0.0) > eps
    assert abs(calibrated_fisheye_param.k4 - 0.0) < eps
    assert abs(calibrated_fisheye_param.k5 - 0.0) < eps
    assert np.all(
        np.asarray(calibrated_fisheye_param.get_intrinsic(3)) == np.asarray(
            fisheye_param.get_intrinsic(3)))
    calibrated_fisheye_param.dump(
        os.path.join(output_dir, 'calibrated_fisheye_param.json'))



# yapf: disable
import logging
from operator import itemgetter
from typing import List, Union

import numpy as np

from xrprimer.data_structure.camera import (
    FisheyeCameraParameter,
    PinholeCameraParameter,
)
from xrprimer.transform.convention.camera import convert_camera_parameter
from xrprimer.utils.log_utils import get_logger
from ..projection.base_projector import BaseProjector

# yapf: enable


# Super class of all triangulators, cannot be built
class BaseTriangulator:
    CAMERA_CONVENTION = 'opencv'
    CAMERA_WORLD2CAM = False

    def __init__(self,
                 camera_parameters: List[Union[PinholeCameraParameter, str]],
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """BaseTriangulator for points triangulation.

        Args:
            camera_parameters (List[Union[PinholeCameraParameter, str]]):
                A list of PinholeCameraParameter, or a list
                of paths to dumped PinholeCameraParameters.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.

        Raises:
            TypeError:
                Some element of camera_parameters is neither
                PinholeCameraParameter nor str.
        """
        self.camera_parameters = []
        self.set_cameras(camera_parameters)
        self.logger = get_logger(logger)

    def set_cameras(
        self, camera_parameters: List[Union[PinholeCameraParameter,
                                            FisheyeCameraParameter]]
    ) -> None:
        """Set cameras for this triangulator.

        Args:
            camera_parameters (List[Union[PinholeCameraParameter, str]]):
                A list of PinholeCameraParameter or FisheyeCameraParameter.

        Raises:
            NotImplementedError:
                Some camera_parameter from camera_parameters
                has a different camera convention from class requirement.
        """
        self.camera_parameters = []
        for input_cam_param in camera_parameters:
            cam_param = input_cam_param.clone()
            if cam_param.world2cam != self.__class__.CAMERA_WORLD2CAM:
                cam_param.inverse_extrinsic()
            if cam_param.convention != self.__class__.CAMERA_CONVENTION:
                cam_param = convert_camera_parameter(
                    cam_param=cam_param, dst=self.__class__.CAMERA_CONVENTION)
            self.camera_parameters.append(cam_param)

    def triangulate(
            self,
            points: Union[np.ndarray, list, tuple],
            points_mask: Union[np.ndarray, list, tuple] = None) -> np.ndarray:
        """Triangulate points with self.camera_parameters.

        Args:
            points (Union[np.ndarray, list, tuple]):
                An ndarray or a nested list of points2d, in shape
                [n_view, n_point 2].
            points_mask (Union[np.ndarray, list, tuple], optional):
                An ndarray or a nested list of mask, in shape
                [n_view, n_point 1].
                If points_mask[index] == 1, points[index] is valid
                for triangulation, else it is ignored.
                If points_mask[index] == np.nan, the whole pair will
                be ignored and not counted by any method.
                Defaults to None.

        Returns:
            np.ndarray:
                An ndarray of points3d, in shape
                [n_point, 3].
        """
        raise NotImplementedError

    def triangulate_single_point(
            self,
            points: Union[np.ndarray, list, tuple],
            points_mask: Union[np.ndarray, list, tuple] = None) -> np.ndarray:
        """Triangulate a single point with self.camera_parameters.

        Args:
            points (Union[np.ndarray, list, tuple]):
                An ndarray or a nested list of points2d, in shape
                [n_view, 2].
            points_mask (Union[np.ndarray, list, tuple], optional):
                An ndarray or a nested list of mask, in shape
                [n_view, 1].
                If points_mask[index] == 1, points[index] is valid
                for triangulation, else it is ignored.
                Defaults to None.

        Returns:
            np.ndarray:
                An ndarray of points3d, in shape
                [3, ].
        """
        raise NotImplementedError

    def get_projector(self) -> BaseProjector:
        projector = BaseProjector(camera_parameters=self.camera_parameters)
        return projector

    def get_reprojection_error(
            self,
            points2d: Union[np.ndarray, list, tuple],
            points3d: Union[np.ndarray, list, tuple],
            points_mask: Union[np.ndarray, list, tuple] = None) -> np.ndarray:
        """Get reprojection error between reprojected points2d and input
        points2d.

        Args:
            points2d (Union[np.ndarray, list, tuple]):
                An ndarray or a nested list of points2d, in shape
                [n_view, n_point, 2].
            points3d (Union[np.ndarray, list, tuple]):
                An ndarray or a nested list of points3d, in shape
                [n_point, 3].
            points_mask (Union[np.ndarray, list, tuple], optional):
                An ndarray or a nested list of mask, in shape
                [n_view, n_point, 1].
                If points_mask[index] == 1, points[index] is valid
                for triangulation, else it is ignored.
                If points_mask[index] == np.nan, the whole pair will
                be ignored and not counted by any method.
                Defaults to None.

        Returns:
            np.ndarray:
                An ndarray in shape [n_view, n_point, 2],
                record offset alone x, y axis of each point2d.
        """
        raise NotImplementedError

    def __get_camera_parameters_slice__(
            self, index: Union[slice, int, list, tuple]) -> list:
        """Slice self.camera_parameters, return a sub-list.

        Args:
            index (Union[slice, int, list, tuple]):
                The index for slicing.

        Returns:
            list:
                A sliced list of self.camera_parameters,
                with selected cameras.
        """
        if isinstance(index, int):
            index = [index]
        if isinstance(index, list) or isinstance(index, tuple):
            new_cam_param_list = itemgetter(*index)(self.camera_parameters)
            if len(index) == 1:
                new_cam_param_list = [
                    new_cam_param_list,
                ]
        else:
            new_cam_param_list = self.camera_parameters[index]
        new_cam_param_list = list(new_cam_param_list)
        return new_cam_param_list

    def __getitem__(self, index: Union[slice, int, list, tuple]):
        """Slice the triangulator by batch dim.

        Args:
            index (Union[slice, int, list, tuple]):
                The index for slicing.

        Returns:
            Triangulator:
                A sliced Triangulator of origin class,
                with selected cameras.
        """
        new_cam_param_list = self.__get_camera_parameters_slice__(index)
        new_triangulator = self.__class__(camera_parameters=new_cam_param_list)
        return new_triangulator



from mmcv.utils import Registry

from .base_calibrator import BaseCalibrator
from .mview_fisheye_calibrator import MviewFisheyeCalibrator
from .mview_pinhole_calibrator import MviewPinholeCalibrator
from .sview_fisheye_calibrator import SviewFisheyeDistortionCalibrator

CALIBRATORS = Registry('calibrator')
CALIBRATORS.register_module(
    name='MviewPinholeCalibrator', module=MviewPinholeCalibrator)
CALIBRATORS.register_module(
    name='MviewFisheyeCalibrator', module=MviewFisheyeCalibrator)
CALIBRATORS.register_module(
    name='SviewFisheyeDistortionCalibrator',
    module=SviewFisheyeDistortionCalibrator)


def build_calibrator(cfg) -> BaseCalibrator:
    """Build calibrator."""
    return CALIBRATORS.build(cfg)



from mmcv.utils import Registry

from .base_projector import BaseProjector
from .opencv_projector import OpencvProjector

PROJECTORS = Registry('projector')
PROJECTORS.register_module(name='OpencvProjector', module=OpencvProjector)


def build_projector(cfg) -> BaseProjector:
    """Build projector."""
    return PROJECTORS.build(cfg)



import copy
import logging
from typing import Union

from xrprimer_cpp.camera import \
    PinholeCameraParameter as PinholeCameraParameter_cpp
from .camera import BaseCameraParameter


class PinholeCameraParameter(PinholeCameraParameter_cpp, BaseCameraParameter):
    ATTR_NAMES = BaseCameraParameter.ATTR_NAMES.copy()

    def __init__(self,
                 K: Union[list, None] = None,
                 R: Union[list, None] = None,
                 T: Union[list, None] = None,
                 name: str = 'default',
                 height: int = 1080,
                 width: int = 1920,
                 world2cam: bool = True,
                 convention: str = 'opencv',
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """A camera parameter class for pinhole camera model. This class has no
        distortion attributes and functions.

        Args:
            K (Union[list, np.ndarray, None], optional):
                Nested list of float32, 4x4 or 3x3 K mat.
                Defaults to None, 4x4 zeros.
            R (Union[list, np.ndarray, None], optional):
                Nested list of float32, 3x3 rotation mat.
                Defaults to None, 3x3 identity.
            T (Union[list, np.ndarray, None], optional):
                List of float32, T vector.
                Defaults to None, zero vector.
            name (str, optional):
                Name of this camera. Defaults to 'default'.
            height (int, optional):
                Height of the image shot by this camera.
                Defaults to 1080.
            width (int, optional):
                Width of the image shot by this camera.
                Defaults to 1920.
            world2cam (bool, optional):
                Whether the R, T transform points from world space
                to camera space. Defaults to True.
            convention (str, optional):
                Convention name of this camera.
                Defaults to 'opencv'.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        PinholeCameraParameter_cpp.__init__(self)
        BaseCameraParameter.__init__(
            self,
            K=K,
            R=R,
            T=T,
            name=name,
            height=height,
            width=width,
            world2cam=world2cam,
            convention=convention,
            logger=logger)

    def clone(self) -> 'PinholeCameraParameter':
        """Clone a new CameraPrameter instance like self.

        Returns:
            PinholeCameraParameter
        """
        new_cam_param = self.__class__(
            K=copy.deepcopy(self.get_intrinsic(k_dim=4)),
            R=copy.deepcopy(self.extrinsic_r),
            T=copy.deepcopy(self.extrinsic_t),
            name=self.name,
            height=self.height,
            width=self.width,
            world2cam=self.world2cam,
            convention=self.convention,
            logger=self.logger)
        return new_cam_param

    def SaveFile(self, filename: str) -> bool:
        """Dump camera name and parameters to a json file.

        Args:
            filename (str):
                Path to the dumped json file.

        Returns:
            bool: True if save succeed.
        """
        return PinholeCameraParameter_cpp.SaveFile(self, filename)

    def LoadFile(self, filename: str) -> bool:
        """Load camera name and parameters from a dumped json file.

        Args:
            filename (str):
                Path to the dumped json file.

        Returns:
            bool: True if load succeed.
        """
        return PinholeCameraParameter_cpp.LoadFile(self, filename)



from xrprimer_cpp import (
    AngleAxisd,
    Pose,
    Quaterniond,
    VectorDouble,
    VectorFloat,
    VectorInt,
    VectorInt64,
    VectorPinholeCameraParameter,
    VectorUint8,
)

__all__ = [
    'AngleAxisd', 'Pose', 'Quaterniond', 'VectorDouble', 'VectorFloat',
    'VectorInt', 'VectorInt64', 'VectorPinholeCameraParameter', 'VectorUint8'
]



from typing import Tuple

import cv2
import numpy as np

from xrprimer.data_structure.camera import (
    FisheyeCameraParameter,
    PinholeCameraParameter,
)
from xrprimer.transform.convention.camera import convert_camera_parameter


def undistort_camera(
        distorted_cam: FisheyeCameraParameter) -> PinholeCameraParameter:
    """Undistort a FisheyeCameraParameter to PinholeCameraParameter.

    Args:
        distorted_cam (FisheyeCameraParameter):
            An instance of FisheyeCameraParameter. Convention will be checked,
            resolution, intrinsic mat
            and distortion coefficients will be used.

    Raises:
        NotImplementedError: Camera convention not supported.

    Returns:
        PinholeCameraParameter:
            Undistorted camera parameter.
    """
    assert isinstance(distorted_cam, FisheyeCameraParameter), \
        'distorted_cam must be an instance of class FisheyeCameraParameter.'
    # prepare input of cv2.undistort
    if distorted_cam.convention != 'opencv':
        distorted_cam = convert_camera_parameter(
            cam_param=distorted_cam, dst='opencv')
    dist_coeff_list = distorted_cam.get_dist_coeff()
    distorted_intrinsic33 = np.array(distorted_cam.get_intrinsic(k_dim=3))
    resolution_wh = np.array([distorted_cam.width, distorted_cam.height])
    # prepare output of cv2.undistort
    corrected_intrinsic33 = np.zeros_like(distorted_intrinsic33)
    corrected_intrinsic33, _ = cv2.getOptimalNewCameraMatrix(
        distorted_intrinsic33, np.array(dist_coeff_list), resolution_wh, 0,
        resolution_wh)
    corrected_cam_param = PinholeCameraParameter(
        K=corrected_intrinsic33,
        R=distorted_cam.get_extrinsic_r(),
        T=distorted_cam.get_extrinsic_t(),
        name=f'undistort_{distorted_cam.name}',
        height=distorted_cam.height,
        width=distorted_cam.width,
        world2cam=distorted_cam.world2cam,
        convention='opencv')
    return corrected_cam_param


def undistort_images(
        distorted_cam: FisheyeCameraParameter,
        image_array: np.ndarray) -> Tuple[PinholeCameraParameter, np.ndarray]:
    """Undistort a FisheyeCameraParameter to PinholeCameraParameter, and
    undistort an array of images shot on a fisheye camera.

    Args:
        distorted_cam (FisheyeCameraParameter):
            An instance of FisheyeCameraParameter. Convention will be checked,
            resolution, intrinsic mat
            and distortion coefficients will be used.
        image_array (np.ndarray):
            An array of images, in shape [n_frame, height, width, n_channel].

    Raises:
        NotImplementedError: Camera convention not supported.

    Returns:
        Tuple[PinholeCameraParameter, np.ndarray]:
            PinholeCameraParameter:
                Undistorted camera parameter.
            np.ndarray:
                Corrected images in the same shape as input.
    """
    # prepare input of cv2.undistort
    if distorted_cam.convention != 'opencv':
        distorted_cam = convert_camera_parameter(
            cam_param=distorted_cam, dst='opencv')
    distorted_intrinsic33 = np.array(distorted_cam.get_intrinsic(k_dim=3))
    dist_coeff_list = distorted_cam.get_dist_coeff()
    dist_coeff_np = np.array(dist_coeff_list)
    corrected_cam_param = undistort_camera(distorted_cam=distorted_cam)
    corrected_intrinsic33 = np.array(
        corrected_cam_param.get_intrinsic(k_dim=3))
    corrected_image_array = np.ones_like(image_array)
    for image_index, image_np in enumerate(image_array):
        corrected_image_array[image_index] = cv2.undistort(
            image_np,
            distorted_intrinsic33,
            dist_coeff_np,
            newCameraMatrix=corrected_intrinsic33)
    return corrected_cam_param, corrected_image_array


def undistort_points(
        distorted_cam: FisheyeCameraParameter,
        points: np.ndarray) -> Tuple[PinholeCameraParameter, np.ndarray]:
    """Undistort a FisheyeCameraParameter to PinholeCameraParameter, and
    undistort an array of points in fisheye camera screen. Parameters and
    points will be casted to np.float64 before operation.

    Args:
        distorted_cam (FisheyeCameraParameter):
            An instance of FisheyeCameraParameter. Convention will be checked,
            resolution, intrinsic mat
            and distortion coefficients will be used.
        points (np.ndarray):
            An array of points, in shape [..., 2], int or float.
            ... could be [n_point, ], [n_frame, n_point, ]
            [n_frame, n_object, n_point, ], etc.

    Raises:
        NotImplementedError: Camera convention not supported.

    Returns:
        Tuple[PinholeCameraParameter, np.ndarray]:
            PinholeCameraParameter:
                Undistorted camera parameter.
            np.ndarray:
                Corrected points location in the same shape as input,
                dtype is np.float64.
    """
    # prepare input of cv2.undistortPoints
    if distorted_cam.convention != 'opencv':
        distorted_cam = convert_camera_parameter(
            cam_param=distorted_cam, dst='opencv')
    distorted_intrinsic33 = np.array(distorted_cam.get_intrinsic(k_dim=3))
    dist_coeff_list = distorted_cam.get_dist_coeff()
    dist_coeff_np = np.array(dist_coeff_list)
    corrected_cam_param = undistort_camera(distorted_cam=distorted_cam)
    corrected_intrinsic33 = np.array(
        corrected_cam_param.get_intrinsic(k_dim=3))
    shape_backup = points.shape
    # opencv expects (n, 1, 2)
    points = points.reshape(-1, 1, 2)
    corrected_points = cv2.undistortPoints(
        points.astype(np.float64),
        cameraMatrix=distorted_intrinsic33,
        distCoeffs=dist_coeff_np,
        P=corrected_intrinsic33)
    corrected_points = corrected_points.reshape(*shape_backup)
    return corrected_cam_param, corrected_points



import glob
import os
import shutil

import cv2
import mmcv
import numpy as np
import pytest

from xrprimer.data_structure.camera import PinholeCameraParameter
from xrprimer.ops.triangulation.builder import build_triangulator  # noqa:E501

input_dir = 'test/data/ops/test_triangulation'
output_dir = 'test/data/output/ops/test_triangulation'


@pytest.fixture(scope='module', autouse=True)
def fixture():
    if os.path.exists(output_dir):
        shutil.rmtree(output_dir)
    os.makedirs(output_dir, exist_ok=False)


def test_opencv_triangulator():
    n_view = len(glob.glob(os.path.join(input_dir, '*.json')))
    kps2d_list = []
    mask_list = []
    for view_idx in range(n_view):
        npz_path = os.path.join(input_dir, f'keypoints_2d_{view_idx:02d}.npz')
        npz_dict = dict(np.load(npz_path, allow_pickle=True))
        kps2d_list.append(npz_dict['keypoints'][0, 0, :, :2])
        mask_list.append(npz_dict['mask'][0, 0, :])
    keypoints2d = np.asarray(kps2d_list)
    keypoints2d_mask = np.asarray(mask_list, dtype=keypoints2d.dtype)
    cam_param_list = []
    for view_idx in range(n_view):
        cam_param_path = os.path.join(input_dir, f'cam_{view_idx:03d}.json')
        cam_param = PinholeCameraParameter()
        cam_param.load(cam_param_path)
        cam_param_list.append(cam_param)
    triangulator_config = dict(
        mmcv.Config.fromfile(
            'config/ops/triangulation/opencv_triangulator.py'))
    triangulator_config['camera_parameters'] = cam_param_list
    triangulator = build_triangulator(triangulator_config)
    # test kp2d np
    keypoints3d = triangulator.triangulate(keypoints2d)
    assert keypoints3d.shape[0] == keypoints2d.shape[1]
    # test kp2d list
    keypoints3d = triangulator.triangulate(keypoints2d.tolist())
    assert keypoints3d.shape[0] == keypoints2d.shape[1]
    # test kp2d tuple
    keypoints3d = triangulator.triangulate(tuple(map(tuple, keypoints2d)))
    assert keypoints3d.shape[0] == keypoints2d.shape[1]
    # test mask np
    points_mask = np.ones_like(keypoints2d[..., 0:1])
    keypoints3d = triangulator.triangulate(
        points=keypoints2d, points_mask=points_mask)
    assert keypoints3d.shape[0] == keypoints2d.shape[1]
    # test mask list
    keypoints3d = triangulator.triangulate(
        points=keypoints2d, points_mask=points_mask.tolist())
    assert keypoints3d.shape[0] == keypoints2d.shape[1]
    # test mask tuple
    keypoints3d = triangulator.triangulate(
        points=keypoints2d, points_mask=tuple(map(tuple, points_mask)))
    assert keypoints3d.shape[0] == keypoints2d.shape[1]
    # test slice
    int_triangulator = triangulator[0]
    assert len(int_triangulator.camera_parameters) == 1
    list_triangulator = triangulator[[0, 1]]
    assert len(list_triangulator.camera_parameters) == 2
    tuple_triangulator = triangulator[(0, 1)]
    assert len(tuple_triangulator.camera_parameters) == 2
    slice_triangulator = triangulator[:2]
    assert len(slice_triangulator.camera_parameters) == 2
    slice_triangulator = triangulator[::2]
    assert len(slice_triangulator.camera_parameters) >= 2
    # test error
    keypoints3d = triangulator.triangulate(keypoints2d)
    error = triangulator.get_reprojection_error(
        points2d=keypoints2d, points3d=keypoints3d)
    assert np.all(error.shape == keypoints2d[..., :2].shape)
    # triangulate and visualize
    keypoints3d = triangulator.triangulate(
        points=keypoints2d, points_mask=np.expand_dims(keypoints2d_mask, -1))
    projector = triangulator.get_projector()
    projected_points = projector.project(keypoints3d)
    for cam_idx in range(n_view):
        canvas = cv2.imread(os.path.join(input_dir, f'{cam_idx:06}.png'))
        valid_idxs = np.where(keypoints2d_mask == 1)
        for point_idx in valid_idxs[1]:
            cv2.circle(
                img=canvas,
                center=projected_points[cam_idx, point_idx].astype(np.int32),
                radius=2,
                color=(0, 0, 255))
        cv2.imwrite(
            filename=os.path.join(output_dir, f'projected_kps_{cam_idx}.jpg'),
            img=canvas)



import os
import shutil

import numpy as np
import pytest

from xrprimer.data_structure.camera import PinholeCameraParameter  # noqa:E501

input_dir = 'test/data/data_structure/camera_parameter'
output_dir = 'test/data/output/data_structure/test_pinhole_camera_parameter'
eps = 1e-4
focal_length_x = 954.5469360351562


@pytest.fixture(scope='module', autouse=True)
def fixture():
    if os.path.exists(output_dir):
        shutil.rmtree(output_dir)
    os.makedirs(output_dir, exist_ok=False)


def test_load():
    input_json_path = os.path.join(input_dir, 'pinhole_param_at_origin.json')
    camera_parameter = PinholeCameraParameter(name='test_LoadFile')
    ret_val = camera_parameter.LoadFile(input_json_path)
    assert ret_val is True
    assert camera_parameter.get_intrinsic(k_dim=4)[0][0] - focal_length_x < eps
    camera_parameter = PinholeCameraParameter(name='test_load')
    camera_parameter.load(input_json_path)
    assert camera_parameter.get_intrinsic(k_dim=4)[0][0] - focal_length_x < eps
    k33_0 = camera_parameter.intrinsic33()
    k33_1 = np.asarray(camera_parameter.get_intrinsic(3))
    assert np.sum(k33_1 - k33_0) < eps
    with pytest.raises(FileNotFoundError):
        camera_parameter.load('/no_file.json')
    with pytest.raises(ValueError):
        camera_parameter.load(
            os.path.join(input_dir, 'omni_param_at_origin.json'))
    camera_parameter = PinholeCameraParameter.fromfile(input_json_path)
    assert camera_parameter.get_intrinsic(k_dim=4)[0][0] - focal_length_x < eps


def test_dump():
    input_json_path = os.path.join(input_dir, 'pinhole_param_at_origin.json')
    output_json_path = os.path.join(output_dir, 'pinhole_param_at_origin.json')
    camera_parameter = PinholeCameraParameter(name='test_dump')
    camera_parameter.load(input_json_path)
    camera_parameter.dump(output_json_path)
    assert os.path.exists(output_json_path)
    os.remove(output_json_path)
    camera_parameter = PinholeCameraParameter(name='test_SaveFile')
    camera_parameter.load(input_json_path)
    camera_parameter.SaveFile(output_json_path)
    assert os.path.exists(output_json_path)
    os.remove(output_json_path)


def test_inverse():
    input_json_path = os.path.join(input_dir,
                                   'pinhole_param_not_at_origin.json')
    camera_parameter = PinholeCameraParameter(name='test_inverse')
    camera_parameter.load(input_json_path)
    origin_world2cam = camera_parameter.world2cam
    origin_intrinsic00 = camera_parameter.get_intrinsic(k_dim=4)[0][0]
    origin_r02 = camera_parameter.get_extrinsic_r()[0][2]
    camera_parameter.inverse_extrinsic()
    assert origin_world2cam is not camera_parameter.world2cam
    assert abs(origin_intrinsic00 -
               camera_parameter.get_intrinsic(k_dim=4)[0][0]) < eps
    assert abs(origin_r02 - camera_parameter.get_extrinsic_r()[0][2]) >= eps
    camera_parameter.inverse_extrinsic()
    assert origin_world2cam is camera_parameter.world2cam
    assert abs(origin_intrinsic00 -
               camera_parameter.get_intrinsic(k_dim=4)[0][0]) < eps
    assert abs(origin_r02 - camera_parameter.get_extrinsic_r()[0][2]) < eps


def test_clone():
    input_json_path = os.path.join(input_dir,
                                   'pinhole_param_not_at_origin.json')
    camera_parameter = PinholeCameraParameter(name='test_clone')
    camera_parameter.load(input_json_path)
    cloned_camera_parameter = camera_parameter.clone()
    assert isinstance(cloned_camera_parameter, PinholeCameraParameter)
    cloned_k = np.asarray(cloned_camera_parameter.get_intrinsic(k_dim=4))
    src_k = np.asarray(camera_parameter.get_intrinsic(k_dim=4))
    assert np.allclose(cloned_k, src_k)
    assert id(cloned_k) != id(src_k)



import logging
from typing import List, Union

import cv2
import numpy as np

from xrprimer.data_structure.camera import (
    FisheyeCameraParameter,
    PinholeCameraParameter,
)
from xrprimer.transform.camera.distortion import undistort_points
from ..projection.opencv_projector import OpencvProjector
from .base_triangulator import BaseTriangulator

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


class OpencvTriangulator(BaseTriangulator):
    CAMERA_CONVENTION = 'opencv'
    CAMERA_WORLD2CAM = True

    def __init__(self,
                 camera_parameters: List[FisheyeCameraParameter],
                 multiview_reduction: Literal['mean', 'median'] = 'mean',
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Triangulator for points triangulation, powered by OpenCV.

        Args:
            camera_parameters (List[FisheyeCameraParameter]):
                A list of FisheyeCameraParameter, or a list
                of PinholeCameraParameter.
            multiview_reduction (Literal['mean', 'median']):
                When more than 2 views are provided, how to
                reduce among view pairs.
                Defaults to mean.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.

        Raises:
            TypeError:
                Some element of camera_parameters is neither
                PinholeCameraParameter nor str.
            NotImplementedError:
                Some camera_parameter from camera_parameters
                has a different camera convention from expectation.
        """
        super().__init__(camera_parameters=camera_parameters, logger=logger)
        self.multiview_reduction = multiview_reduction

    def triangulate(
            self,
            points: Union[np.ndarray, list, tuple],
            points_mask: Union[np.ndarray, list, tuple] = None) -> np.ndarray:
        """Triangulate points with self.camera_parameters.

        Args:
            points (Union[np.ndarray, list, tuple]):
                An ndarray or a nested list of points2d, in shape
                [n_view, n_point 2].
            points_mask (Union[np.ndarray, list, tuple], optional):
                An ndarray or a nested list of mask, in shape
                [n_view, n_point 1].
                If points_mask[index] == 1, points[index] is valid
                for triangulation, else it is ignored.
                If points_mask[index] == np.nan, the whole pair will
                be ignored and not counted by any method.
                Defaults to None.

        Returns:
            np.ndarray:
                An ndarray of points3d, in shape
                [n_point, 3].
        """
        assert len(points) == len(self.camera_parameters)
        points = np.array(points)
        undistorted_cam_list = []
        for view_idx, view_cam in enumerate(self.camera_parameters):
            if isinstance(view_cam, FisheyeCameraParameter):
                undistorted_cam, points[view_idx, ...] = undistort_points(
                    distorted_cam=view_cam, points=points[view_idx, ...])
                undistorted_cam_list.append(undistorted_cam)
            else:
                undistorted_cam_list.append(view_cam)
        triangulation_mat = self.__class__.prepare_triangulation_mat(
            undistorted_cam_list)
        if points_mask is not None:
            points_mask = np.array(points_mask)
        else:
            points_mask = np.ones_like(points[:, :, :1])
        n_view = len(self.camera_parameters)
        n_point = points.shape[1]
        n_pair = int(n_view * (n_view - 1) / 2)
        triangulation_results = np.zeros(shape=(n_pair, n_point, 3))
        triangulation_result_mask = np.zeros(shape=(n_pair, n_point, 1))
        pair_count = 0
        for view_index_0 in range(n_view):
            for view_index_1 in range(view_index_0 + 1, n_view, 1):
                # if mask of one point is 0 or nan,
                # the pair fails(masked by nan)
                pair_mask = points_mask[view_index_0, ...] * \
                    points_mask[view_index_1, ...]
                pair_mask[pair_mask == 0] = np.nan
                point4d_hom = cv2.triangulatePoints(
                    projMatr1=triangulation_mat[view_index_0],
                    projMatr2=triangulation_mat[view_index_1],
                    projPoints1=points[view_index_0, :, :2].transpose(1, 0),
                    projPoints2=points[view_index_1, :, :2].transpose(1, 0))
                # if 0, invalid, else divide it
                dividend = point4d_hom[3, :]
                pair_mask[dividend == 0, :] = np.nan
                # avoid zero-division
                dividend[dividend == 0] = 1
                point3d = point4d_hom[:3] / dividend
                triangulation_results[pair_count, :, :] = point3d.transpose(
                    1, 0)
                triangulation_result_mask[pair_count, :, :] = pair_mask
                pair_count += 1
        triangulation_result_mask = np.repeat(
            triangulation_result_mask, 3, axis=2)
        triangulation_results = triangulation_results * \
            triangulation_result_mask
        # get mean location among pairs ignoring nan
        if self.multiview_reduction == 'mean':
            points3d = np.nanmean(
                triangulation_results, axis=0, keepdims=False)
        elif self.multiview_reduction == 'median':
            points3d = np.nanmedian(
                triangulation_results, axis=0, keepdims=False)
        else:
            self.logger.error(
                f'Wrong reduction_method: {self.multiview_reduction}')
            raise ValueError
        points3d[np.isnan(points3d)] = 0.0
        return points3d

    def triangulate_single_point(
            self,
            points: Union[np.ndarray, list, tuple],
            points_mask: Union[np.ndarray, list, tuple] = None) -> np.ndarray:
        """Triangulate a single point with self.camera_parameters.

        Args:
            points (Union[np.ndarray, list, tuple]):
                An ndarray or a nested list of points2d, in shape
                [n_view, 2].
            points_mask (Union[np.ndarray, list, tuple], optional):
                An ndarray or a nested list of mask, in shape
                [n_view, 1].
                If points_mask[index] == 1, points[index] is valid
                for triangulation, else it is ignored.
                Defaults to None.

        Returns:
            np.ndarray:
                An ndarray of points3d, in shape
                [3, ].
        """
        assert len(points) == len(self.camera_parameters)
        points = np.array(points)
        undistorted_cam_list = []
        for view_idx, view_cam in enumerate(self.camera_parameters):
            if isinstance(view_cam, FisheyeCameraParameter):
                undistorted_cam, points[view_idx, ...] = undistort_points(
                    distorted_cam=view_cam, points=points[view_idx, ...])
                undistorted_cam_list.append(undistorted_cam)
            else:
                undistorted_cam_list.append(view_cam)
        triangulation_mat = self.__class__.prepare_triangulation_mat(
            undistorted_cam_list)
        if points_mask is not None:
            points_mask = np.array(points_mask)
        else:
            points_mask = np.ones_like(points[:, :1])
        n_view = len(self.camera_parameters)
        triangulation_results = []
        for view_index_0 in range(n_view):
            if points_mask[view_index_0, 0] != 1:
                continue
            for view_index_1 in range(view_index_0 + 1, n_view, 1):
                if points_mask[view_index_1, 0] != 1:
                    continue
                point4d_hom = cv2.triangulatePoints(
                    projMatr1=triangulation_mat[view_index_0],
                    projMatr2=triangulation_mat[view_index_1],
                    projPoints1=points[view_index_0],
                    projPoints2=points[view_index_1])
                # if 0, invalid, else divide it
                dividend = point4d_hom[3, :]
                if dividend == 0:
                    continue
                point3d = point4d_hom[:3] / dividend
                triangulation_results.append(np.squeeze(point3d, axis=1))
        point3d = np.zeros(shape=[
            3,
        ])
        if len(triangulation_results) > 0:
            if self.multiview_reduction == 'mean':
                point3d = np.mean(
                    np.array(triangulation_results), axis=0, keepdims=False)
            elif self.multiview_reduction == 'median':
                point3d = np.median(
                    np.array(triangulation_results), axis=0, keepdims=False)
            else:
                self.logger.error(
                    f'Wrong reduction_method: {self.multiview_reduction}')
                raise ValueError
        return point3d

    def get_reprojection_error(
            self,
            points2d: Union[np.ndarray, list, tuple],
            points3d: Union[np.ndarray, list, tuple],
            points_mask: Union[np.ndarray, list, tuple] = None) -> np.ndarray:
        """Get reprojection error between reprojected points2d and input
        points2d.

        Args:
            points2d (Union[np.ndarray, list, tuple]):
                An ndarray or a nested list of points2d, in shape
                [n_view, n_point, 2].
            points3d (Union[np.ndarray, list, tuple]):
                An ndarray or a nested list of points3d, in shape
                [n_point, 3].
            points_mask (Union[np.ndarray, list, tuple], optional):
                An ndarray or a nested list of mask, in shape
                [n_view, n_point, 1].
                If points_mask[index] == 1, points[index] is valid
                for triangulation, else it is ignored.
                If points_mask[index] == np.nan, the whole pair will
                be ignored and not counted by any method.
                Defaults to None.

        Returns:
            np.ndarray:
                An ndarray in shape [n_view, n_point, 2],
                record offset alone x, y axis of each point2d.
        """
        projector = self.get_projector()
        points3d = np.array(points3d).reshape(-1, 3)
        points2d_shape_backup = points2d.shape
        n_view = points2d_shape_backup[0]
        points2d = np.array(points2d).reshape(n_view, -1, 2)
        if points_mask is None:
            points_mask = np.ones_like(points2d[..., :1])
        points_mask = np.array(points_mask).reshape(n_view, -1, 1)
        projected_points = projector.project(points3d)
        projected_error = projected_points - points2d
        points_mask = np.repeat(points_mask, 2, axis=2)
        projected_error = projected_error * points_mask
        return projected_error

    def get_projector(self) -> OpencvProjector:
        projector = OpencvProjector(camera_parameters=self.camera_parameters)
        return projector

    def __getitem__(self, index: Union[slice, int, list, tuple]):
        """Slice the triangulator by batch dim.

        Args:
            index (Union[slice, int, list, tuple]):
                The index for slicing.

        Returns:
            Triangulator:
                A sliced Triangulator of origin class,
                with selected cameras.
        """
        new_cam_param_list = self.__get_camera_parameters_slice__(index)
        new_triangulator = self.__class__(
            camera_parameters=new_cam_param_list,
            multiview_reduction=self.multiview_reduction)
        return new_triangulator

    @classmethod
    def prepare_triangulation_mat(
            cls,
            camera_parameters: List[PinholeCameraParameter]) -> np.ndarray:
        """Prepare projection matrix for triangulation. According to opencv,

        ProjectionMatrix = [intrinsic33] * [extrinsic_r|extrinsic_t]

        Args:
            camera_parameters (List[PinholeCameraParameter]):
                A list of pinhole camera parameters.

        Returns:
            np.ndarray:
                The projection matrix in shape
                [n_camera, 3, 4].
        """
        triangulation_mat = np.zeros(shape=(len(camera_parameters), 3, 4))
        for camera_index in range(len(camera_parameters)):
            triangulation_mat[camera_index, :, :3] = np.array(
                camera_parameters[camera_index].get_extrinsic_r()).reshape(
                    3, 3)
            triangulation_mat[camera_index, :, 3:] = np.array(
                camera_parameters[camera_index].get_extrinsic_t()).reshape(
                    3, 1)
            triangulation_mat[camera_index] = np.matmul(
                np.array(
                    camera_parameters[camera_index].get_intrinsic(k_dim=3)),
                triangulation_mat[camera_index])
        return triangulation_mat



import logging
from typing import List, Union

import cv2
import numpy as np

from xrprimer.data_structure.camera import FisheyeCameraParameter
from .base_projector import BaseProjector


class OpencvProjector(BaseProjector):
    CAMERA_CONVENTION = 'opencv'
    CAMERA_WORLD2CAM = True

    def __init__(self,
                 camera_parameters: List[FisheyeCameraParameter],
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Projector for points projection, powered by OpenCV.

        Args:
            camera_parameters (List[FisheyeCameraParameter]):
                A list of FisheyeCameraParameter.
        """
        BaseProjector.__init__(self, camera_parameters, logger=logger)

    def project(
            self,
            points: Union[np.ndarray, list, tuple],
            points_mask: Union[np.ndarray, list, tuple] = None) -> np.ndarray:
        """Project points with self.camera_parameters.

        Args:
            points (Union[np.ndarray, list, tuple]):
                An ndarray or a nested list of points3d, in shape
                [n_point, 3].
            points_mask (Union[np.ndarray, list, tuple], optional):
                An ndarray or a nested list of mask, in shape
                [n_point, 1].
                If points_mask[index] == 1, points[index] is valid
                for projection, else it is ignored.
                Defaults to None.

        Returns:
            np.ndarray:
                An ndarray of points2d, in shape
                [n_view, n_point, 2].
        """
        distortion_list = self.__prepare_dist_coeff__()
        points3d = np.array(points, dtype=np.float64).reshape(-1, 3)
        n_view = len(self.camera_parameters)
        n_point = points3d.shape[0]
        points2d = np.zeros(shape=[n_view, n_point, 2], dtype=points3d.dtype)
        points_mask = np.array(points_mask).reshape(-1) \
            if points_mask is not None \
            else np.ones(shape=[n_point, ], dtype=np.uint8)
        valid_idxs = np.where(points_mask == 1)
        for camera_index in range(len(self.camera_parameters)):
            cam_param = self.camera_parameters[camera_index]
            r_mat = np.array(cam_param.get_extrinsic_r(), dtype=points3d.dtype)
            t_vec = np.array(cam_param.get_extrinsic_t(), dtype=points3d.dtype)
            k_mat = np.array(cam_param.get_intrinsic(3), dtype=points3d.dtype)
            dist_coeffs = np.array(
                distortion_list[camera_index], dtype=points3d.dtype)
            projected_points, _ = cv2.projectPoints(
                objectPoints=points3d[valid_idxs[0], :],
                rvec=r_mat,
                tvec=t_vec,
                cameraMatrix=k_mat,
                distCoeffs=dist_coeffs)
            projected_points = projected_points.reshape(-1, 2)
            points2d[camera_index, valid_idxs[0], :] = projected_points
        return points2d

    def project_single_point(
            self, points: Union[np.ndarray, list, tuple]) -> np.ndarray:
        """Project a single point with self.camera_parameters.

        Args:
            points (Union[np.ndarray, list, tuple]):
                An ndarray or a list of points3d, in shape
                [3].

        Returns:
            np.ndarray:
                An ndarray of points2d, in shape
                [n_view, 2].
        """
        points3d = np.array(points).reshape(1, 3)
        return np.squeeze(self.project(points3d), axis=1)

    def __prepare_dist_coeff__(self) -> list:
        """Prepare distortion argument for opencv.

        Returns:
            list: list of distCoeffs, len(list) == n_camera.
        """
        distortion_list = []
        for cam_param in self.camera_parameters:
            if hasattr(cam_param, 'k1'):
                dist_coeffs = np.array([
                    cam_param.k1, cam_param.k2, cam_param.p1, cam_param.p2,
                    cam_param.k3, cam_param.k4, cam_param.k5, cam_param.k6
                ])
            else:
                dist_coeffs = np.zeros(shape=[
                    8,
                ])
            distortion_list.append(dist_coeffs)
        return distortion_list



import os
import re
import sys

import xrprimer_cpp as xr

calib_config = """
    {
      "chessboard_width": 6,
      "chessboard_height": 7,
      "chessboard_square_size": 100
    }
"""

image_folder = 'test/data/calib_pinhole_camera/input/images/'
camera_folder = 'test/data/calib_pinhole_camera/input/config/'

if __name__ == '__main__':

    if len(sys.argv) >= 3:
        image_folder = sys.argv[1]
        camera_folder = sys.argv[2]
    print('image folder: ', image_folder)
    print('camera folder: ', camera_folder)

    pinhole_cam_list = []
    # pinhole_list = []
    for root, dirs, files in os.walk(camera_folder):
        camera_cnt = len(list(files))
        pinhole_list = [''] * camera_cnt
        pinhole_cam_list = [
            xr.camera.PinholeCameraParameter() for i in range(camera_cnt)
        ]

        for file in files:
            camera_idx = int(os.path.splitext(file.split('_')[1])[0])
            # pinhole_list[camera_idx] = os.path.join(root, file)
            cam_path = os.path.join(root, file)
            pinhole_cam_list[camera_idx].LoadFile(cam_path)
            # pinhole_cam_list[camera_idx].SaveFile(str(camera_idx) + ".json")

    max_img_idx = 0
    frames = []
    for root, dirs, files in os.walk(image_folder):
        for file in files:
            img = os.path.join(root, file)
            matched = re.match(
                r'img(?P<img_index>\d+)_cam(?P<camera_index>\d+).jpg', file)
            if max_img_idx < int(matched.group('img_index')):
                max_img_idx = int(matched.group('img_index'))

        frames = [[]] * (max_img_idx + 1)
        for x in range(max_img_idx + 1):
            frames[x] = [''] * 10

        for file in files:
            img = os.path.join(root, file)
            matched = re.match(
                r'img(?P<img_index>\d+)_cam(?P<camera_index>\d+).jpg', file)
            frames[int(matched.group('img_index'))][int(
                matched.group('camera_index'))] = img

    vc = xr.VectorPinholeCameraParameter(pinhole_cam_list)
    xr.calibrator.CalibrateMultiPinholeCamera(calib_config, frames, vc)
    print(vc[1].extrinsic_r)

    for idx, pin in enumerate(vc):
        out_path = f'outputs/{idx}.json'
        os.makedirs(os.path.dirname(out_path), exist_ok=True)
        print(f'save file: {out_path}')
        pin.SaveFile(out_path)



from typing import Union

import numpy as np

from xrprimer.data_structure.camera import (
    FisheyeCameraParameter,
    PinholeCameraParameter,
)


def rotate_camera(
    cam_param: Union[PinholeCameraParameter,
                     FisheyeCameraParameter], rotation_mat: np.ndarray
) -> Union[PinholeCameraParameter, FisheyeCameraParameter]:
    """Apply rotation to a camera parameter.

    Args:
        cam_param (Union[
                PinholeCameraParameter,
                FisheyeCameraParameter]):
            The camera to rotate.
        rotation_mat (np.ndarray):
            Rotation matrix defined in world space,
            shape [3, 3].

    Returns:
        Union[PinholeCameraParameter, FisheyeCameraParameter]:
            Rotated camera in same type and extrinsic direction
            like the input camera.
    """
    input_w2c = cam_param.world2cam
    ret_cam_param = cam_param.clone()
    if input_w2c:
        ret_cam_param.inverse_extrinsic()
    cam_loc = np.asarray(ret_cam_param.get_extrinsic_t())
    cam_loc = np.matmul(rotation_mat, cam_loc)
    c2w_rot = np.asarray(ret_cam_param.get_extrinsic_r())
    c2w_rot = np.matmul(rotation_mat, c2w_rot)
    ret_cam_param.set_KRT(R=c2w_rot, T=cam_loc, world2cam=False)
    if ret_cam_param.world2cam != input_w2c:
        ret_cam_param.inverse_extrinsic()
    return ret_cam_param


def translate_camera(
    cam_param: Union[PinholeCameraParameter,
                     FisheyeCameraParameter], translation: np.ndarray
) -> Union[PinholeCameraParameter, FisheyeCameraParameter]:
    """Apply the translation to a camera parameter.

    Args:
        cam_param (Union[
                PinholeCameraParameter,
                FisheyeCameraParameter]):
            The camera to rotate.
        translation (np.ndarray):
            Translation vector defined in world space,
            shape [3,].

    Returns:
        Union[PinholeCameraParameter, FisheyeCameraParameter]:
            Translated camera in same type and extrinsic direction
            like the input camera.
    """
    input_w2c = cam_param.world2cam
    ret_cam_param = cam_param.clone()
    if input_w2c:
        ret_cam_param.inverse_extrinsic()
    cam_loc = np.asarray(ret_cam_param.get_extrinsic_t())
    cam_loc += translation
    ret_cam_param.set_KRT(T=cam_loc, world2cam=False)
    if ret_cam_param.world2cam != input_w2c:
        ret_cam_param.inverse_extrinsic()
    return ret_cam_param



import numpy as np

from xrprimer_cpp import VectorDouble
from xrprimer_cpp.ops import prior_guided_pnp


def pose_estimation():
    m, n = 5, 100
    correspondences = np.random.random([m, n]).astype('float32')
    point2ds = correspondences[0:2, :].copy()
    point3ds = correspondences[2:5, :].copy()
    priors = np.ones([1, n], dtype='float32')
    params = VectorDouble([1465., 1465., 955., 689.])
    camera_config = {'model_name': 'PINHOLE', 'params': params}
    ransac_config = {
        'error_thres': 12,
        'inlier_ratio': 0.01,
        'confidence': 0.9999,
        'max_iter': 100000,
        'local_optimal': True
    }
    return prior_guided_pnp(point2ds, point3ds, priors, camera_config,
                            ransac_config)


if __name__ == '__main__':
    print(pose_estimation())



import glob
import json
import os
import shutil

import mmcv
import numpy as np
import pytest

from xrprimer.calibration.builder import build_calibrator
from xrprimer.data_structure.camera import (
    FisheyeCameraParameter,
    PinholeCameraParameter,
)
from xrprimer.utils.path_utils import Existence, check_path_existence

input_dir = 'test/data/calib_pinhole_camera/input'
output_dir = 'test/data/output/calibration/test_mview_extrinsic_calibrator'


@pytest.fixture(scope='module', autouse=True)
def fixture():
    if os.path.exists(output_dir):
        shutil.rmtree(output_dir)
    os.makedirs(output_dir, exist_ok=False)


def get_frame_list(n_view: int) -> list:
    n_frame = len(glob.glob(os.path.join(input_dir, 'images', '*_cam00.jpg')))
    mframe_list = []
    for frame_idx in range(n_frame):
        mview_list = []
        for view_idx in range(n_view):
            img_path = os.path.join(
                input_dir, 'images',
                f'img{frame_idx:04d}_cam{view_idx:02d}.jpg')
            if check_path_existence(img_path) == Existence.FileExist:
                mview_list.append(img_path)
            else:
                mview_list.append('')
        mframe_list.append(mview_list)
    return mframe_list


def test_mview_pinhole_calibrator():
    # init pinhole parameters with intrinsic
    init_param_dir = os.path.join(input_dir, 'config')
    file_names = sorted(glob.glob(os.path.join(init_param_dir, '*.json')))
    pinhole_list = []
    for cam_idx, file_path in enumerate(file_names):
        with open(file_path, 'r') as f_read:
            param_dict = json.load(f_read)
        init_k = np.asarray(param_dict['intrinsic'])
        pinhole_param = PinholeCameraParameter(
            K=init_k, name=f'pinhole_{cam_idx:02d}', convention='opencv')
        pinhole_list.append(pinhole_param)
    n_view = len(pinhole_list)
    mframe_list = get_frame_list(n_view)
    # test dict config
    calibrator_config = dict(
        mmcv.Config.fromfile('config/calibration/' +
                             'mview_pinhole_calibrator.py'))
    calibrator = build_calibrator(calibrator_config)
    pinhole_list = calibrator.calibrate(mframe_list, pinhole_list)
    for pinhole_param in pinhole_list:
        pinhole_param.dump(
            os.path.join(output_dir, f'{pinhole_param.name}.json'))


def test_mview_fisheye_calibrator():
    # init pinhole parameters with intrinsic
    init_param_dir = os.path.join(input_dir, 'config')
    file_names = sorted(glob.glob(os.path.join(init_param_dir, '*.json')))
    fisheye_list = []
    for cam_idx, file_path in enumerate(file_names):
        with open(file_path, 'r') as f_read:
            param_dict = json.load(f_read)
        init_k = np.asarray(param_dict['intrinsic'])
        fisheye_param = FisheyeCameraParameter(
            K=init_k, name=f'fisheye_{cam_idx:02d}', convention='opencv')
        fisheye_param.set_dist_coeff([
            1e-4,
        ] * 6, [
            1e-4,
        ] * 2)
        fisheye_list.append(fisheye_param)
    n_view = len(fisheye_list)
    mframe_list = get_frame_list(n_view)
    # test dict config
    calibrator_config = dict(
        mmcv.Config.fromfile('config/calibration/' +
                             'mview_fisheye_calibrator.py'))
    # calibrate only extrinsic
    calibrator_config['calibrate_intrinsic'] = False
    calibrator_config['calibrate_distortion'] = False
    calibrator_config['calibrate_extrinsic'] = True
    calibrator_config['work_dir'] = os.path.join(
        output_dir, 'mview_fisheye_extrinsic_calibrator_temp')
    calibrator = build_calibrator(calibrator_config)
    fisheye_list = calibrator.calibrate(mframe_list, fisheye_list)
    for fisheye_param in fisheye_list:
        fisheye_param.dump(
            os.path.join(output_dir, f'fisheye_calibext_{cam_idx:02d}.json'))
    # calibrate only distortion
    calibrator_config['calibrate_intrinsic'] = False
    calibrator_config['calibrate_distortion'] = True
    calibrator_config['calibrate_extrinsic'] = False
    calibrator_config['work_dir'] = os.path.join(
        output_dir, 'mview_fisheye_extrinsic_calibrator_temp')
    calibrator = build_calibrator(calibrator_config)
    fisheye_list = calibrator.calibrate(mframe_list, fisheye_list)
    for fisheye_param in fisheye_list:
        fisheye_param.dump(
            os.path.join(output_dir, f'fisheye_calibdist_{cam_idx:02d}.json'))
    # calibrate distortion and extrinsic
    calibrator_config['calibrate_intrinsic'] = False
    calibrator_config['calibrate_distortion'] = True
    calibrator_config['calibrate_extrinsic'] = True
    calibrator_config['work_dir'] = os.path.join(
        output_dir, 'mview_fisheye_extrinsic_calibrator_temp')
    calibrator = build_calibrator(calibrator_config)
    fisheye_list = calibrator.calibrate(mframe_list, fisheye_list)
    for fisheye_param in fisheye_list:
        fisheye_param.dump(
            os.path.join(output_dir,
                         f'fisheye_calibdistext_{cam_idx:02d}.json'))



import numpy as np


def upgrade_k_3x3(k: np.ndarray, is_perspective: bool = True) -> np.ndarray:
    """Convert opencv 3x3 intrinsic matrix to 4x4.

    Args:
        K (np.ndarray):
            Input 3x3 intrinsic matrix, left mm defined.
            [[fx,   0,   px],
             [0,   fy,   py],
             [0,    0,   1]]
        is_perspective (bool, optional): whether is perspective projection.
            Defaults to True.

    Returns:
        np.ndarray:
            Output intrinsic matrix.
            for perspective:
                [[fx,   0,    px,   0],
                [0,   fy,    py,   0],
                [0,    0,    0,    1],
                [0,    0,    1,    0]]

            for orthographics:
                [[fx,   0,    0,   px],
                [0,   fy,    0,   py],
                [0,    0,    1,    0],
                [0,    0,    0,    1]]
    """
    k_batch = k.reshape(-1, 3, 3)
    if is_perspective:
        k_ret = np.zeros((k_batch.shape[0], 4, 4), dtype=k_batch.dtype)
        k_ret[:, :2, :3] = k_batch[:, :2, :3]
        k_ret[:, 3, 2] = 1
        k_ret[:, 2, 3] = 1
    else:
        k_ret = np.zeros((k_batch.shape[0], 4, 4), dtype=k_batch.dtype)
        k_ret[:, :2, :2] = k_batch[:, :2, :2]
        k_ret[:, :2, 3:] = k_batch[:, :2, 2:]
        k_ret[:, 2, 2] = 1
        k_ret[:, 3, 3] = 1
    ret_shape = [4, 4]
    for dim_index in range(k.ndim - 3, -1, -1):
        ret_shape.insert(0, k.shape[dim_index])
    return k_ret.reshape(*ret_shape)


def downgrade_k_4x4(k: np.ndarray) -> np.ndarray:
    """Convert opencv 4x4 intrinsic matrix to 3x3.

    Args:
        K (np.ndarray):
            Input 4x4 intrinsic matrix, left mm defined.

    Returns:
        np.ndarray:
            Output 3x3 intrinsic matrix, left mm defined.
            [[fx,   0,   px],
             [0,   fy,   py],
             [0,    0,   1]]
    """
    k_batch = k.reshape(-1, 4, 4)
    is_perspective = (k_batch[0, 2, 3] == k_batch[0, 3, 2])
    if is_perspective:
        k_ret = np.zeros((k_batch.shape[0], 3, 3), dtype=k_batch.dtype)
        k_ret[:, :2, :3] = k_batch[:, :2, :3]
        k_ret[:, 2, 2] = 1
    else:
        k_ret = np.zeros((k_batch.shape[0], 3, 3), dtype=k_batch.dtype)
        k_ret[:, :2, :2] = k_batch[:, :2, :2]
        k_ret[:, :2, 2:3] = k_batch[:, :2, 3:4]
        k_ret[:, 2, 2] = 1
    ret_shape = [3, 3]
    for dim_index in range(k.ndim - 3, -1, -1):
        ret_shape.insert(0, k.shape[dim_index])
    return k_ret.reshape(*ret_shape)



type = 'MviewFisheyeCalibrator'
work_dir = './temp'
chessboard_width = 6
chessboard_height = 7
chessboard_square_size = 100
calibrate_intrinsic = False
calibrate_distortion = False
calibrate_extrinsic = True



import numpy as np

from xrprimer.transform.image.color import bgr2rgb


def test_bgr2rgb():
    # test numpy
    # one img
    rgb_image = np.zeros(shape=(3, 640, 480))
    rgb_image[2, ...] = 2
    assert rgb_image[2, 0, 0] == 2
    bgr_image = bgr2rgb(rgb_image, color_dim=0)
    assert bgr_image[0, 0, 0] == 2
    assert bgr_image[2, 0, 0] == 0
    # pytorch batch like
    rgb_image = np.zeros(shape=(2, 3, 640, 480))
    rgb_image[:, 2, ...] = 2
    assert rgb_image[0, 2, 0, 0] == 2
    bgr_image = bgr2rgb(rgb_image, color_dim=1)
    assert bgr_image[0, 0, 0, 0] == 2
    assert bgr_image[0, 2, 0, 0] == 0
    # opencv video like
    rgb_image = np.zeros(shape=(2, 640, 480, 3))
    rgb_image[..., 2] = 2
    assert rgb_image[0, 0, 0, 2] == 2
    bgr_image = bgr2rgb(rgb_image, color_dim=-1)
    assert bgr_image[0, 0, 0, 0] == 2
    assert bgr_image[0, 0, 0, 2] == 0



import copy
import logging
from typing import Union

from xrprimer_cpp.camera import \
    FisheyeCameraParameter as FisheyeCameraParameter_cpp
from .pinhole_camera import PinholeCameraParameter


class FisheyeCameraParameter(FisheyeCameraParameter_cpp,
                             PinholeCameraParameter):
    ATTR_NAMES = PinholeCameraParameter.ATTR_NAMES.copy() + \
        ['k1', 'k2', 'k3', 'k4', 'k5', 'k6', 'p1', 'p2']

    def __init__(self,
                 K: Union[list, None] = None,
                 R: Union[list, None] = None,
                 T: Union[list, None] = None,
                 name: str = 'default',
                 height: int = 1080,
                 width: int = 1920,
                 world2cam: bool = True,
                 convention: str = 'opencv',
                 dist_coeff_k: list = [],
                 dist_coeff_p: list = [],
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """A camera parameter class for fisheye camera model. Distortion
        coefficients are k1, k2, ..., k6, p1, p2.

        Args:
            K (Union[list, np.ndarray, None], optional):
                Nested list of float32, 4x4 or 3x3 K mat.
                Defaults to None, 4x4 zeros.
            R (Union[list, np.ndarray, None], optional):
                Nested list of float32, 3x3 rotation mat.
                Defaults to None, 3x3 identity.
            T (Union[list, np.ndarray, None], optional):
                List of float32, T vector.
                Defaults to None, zero vector.
            name (str, optional):
                Name of this camera. Defaults to 'default'.
            height (int, optional):
                Height of the image shot by this camera.
                Defaults to 1080.
            width (int, optional):
                Width of the image shot by this camera.
                Defaults to 1920.
            world2cam (bool, optional):
                Whether the R, T transform points from world space
                to camera space. Defaults to True.
            convention (str, optional):
                Convention name of this camera.
                Defaults to 'opencv'.
            dist_coeff_k (list, optional):
                List of float. [k1, k2, k3, k4, k5, k6].
                When length of list is n and n<6,
                only the first n coefficients will be set. Defaults to [].
            dist_coeff_p (list, optional):
                List of float. [p1, p2].
                To set only p1, pass [p1]. Defaults to [].
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        FisheyeCameraParameter_cpp.__init__(self)
        PinholeCameraParameter.__init__(
            self,
            K=K,
            R=R,
            T=T,
            name=name,
            height=height,
            width=width,
            world2cam=world2cam,
            convention=convention,
            logger=logger)
        self.set_dist_coeff(
            dist_coeff_k=dist_coeff_k, dist_coeff_p=dist_coeff_p)

    def set_dist_coeff(self, dist_coeff_k: list, dist_coeff_p: list) -> None:
        """Set distortion coefficients from list.

        Args:
            dist_coeff_k (list):
                List of float. [k1, k2, k3, k4, k5, k6].
                When length of list is n and n<6,
                only the first n coefficients will be set.
            dist_coeff_p (list):
                List of float. [p1, p2].
                To set only p1, pass [p1].
        """
        assert len(dist_coeff_k) <= 6
        assert len(dist_coeff_p) <= 2
        for k_index, k_value in enumerate(dist_coeff_k):
            setattr(self, f'k{k_index+1}', k_value)
        for p_index, p_value in enumerate(dist_coeff_p):
            setattr(self, f'p{p_index+1}', p_value)

    def get_dist_coeff(self) -> list:
        """Get distortion coefficients in self.convention.

        Raises:
            NotImplementedError: convention not supported.

        Returns:
            list: A list of distortion coefficients, in a
            turn defined by self.convention.
        """
        dist_coeff_list = []
        if self.convention == 'opencv':
            dist_coeff_names = ['k1', 'k2', 'p1', 'p2', 'k3', 'k4', 'k5', 'k6']
            for coeff_name in dist_coeff_names:
                dist_coeff_list.append(getattr(self, coeff_name, 0.0))
        else:
            self.logger.error(f'Distortion for camera in {self.convention}' +
                              ' has not been supported .')
            raise NotImplementedError
        return dist_coeff_list

    def clone(self) -> 'FisheyeCameraParameter':
        """Clone a new CameraPrameter instance like self.

        Returns:
            FisheyeCameraParameter
        """
        new_cam_param = self.__class__(
            K=copy.deepcopy(self.get_intrinsic(k_dim=4)),
            R=copy.deepcopy(self.extrinsic_r),
            T=copy.deepcopy(self.extrinsic_t),
            name=self.name,
            height=self.height,
            width=self.width,
            world2cam=self.world2cam,
            convention=self.convention,
            logger=self.logger)
        attr_name_list = [f'k{k_index}' for k_index in range(1, 7, 1)]
        attr_name_list += [f'p{p_index}' for p_index in range(1, 3, 1)]
        for attr_name in attr_name_list:
            attr_value = getattr(self, attr_name)
            setattr(new_cam_param, attr_name, attr_value)
        return new_cam_param

    def SaveFile(self, filename: str) -> bool:
        """Dump camera name and parameters to a json file.

        Args:
            filename (str):
                Path to the dumped json file.

        Returns:
            bool: True if save succeed.
        """
        return FisheyeCameraParameter_cpp.SaveFile(self, filename)

    def LoadFile(self, filename: str) -> bool:
        """Load camera name and parameters from a dumped json file.

        Args:
            filename (str):
                Path to the dumped json file.

        Returns:
            bool: True if load succeed.
        """
        return FisheyeCameraParameter_cpp.LoadFile(self, filename)



from mmcv.utils import Registry

from .base_triangulator import BaseTriangulator
from .opencv_triangulator import OpencvTriangulator

TRIANGULATORS = Registry('triangulator')
TRIANGULATORS.register_module(
    name='OpencvTriangulator', module=OpencvTriangulator)


def build_triangulator(cfg) -> BaseTriangulator:
    """Build triangulator."""
    return TRIANGULATORS.build(cfg)



import os
import shutil

import cv2
import mmcv
import numpy as np
import pytest

from xrprimer.data_structure.camera import PinholeCameraParameter
from xrprimer.ops.projection.builder import build_projector
from xrprimer.transform.camera.extrinsic import rotate_camera, translate_camera

input_dir = 'test/data/transform/camera/test_extrinsic'
output_dir = 'test/data/output/transform/camera/test_extrinsic'


@pytest.fixture(scope='module', autouse=True)
def fixture():
    if os.path.exists(output_dir):
        shutil.rmtree(output_dir)
    os.makedirs(output_dir, exist_ok=False)


def test_rotate_camera():
    src_cam_param = PinholeCameraParameter.fromfile(
        os.path.join(input_dir, 'cam_006.json'))
    flip_rot_mat = np.array([
        [1, 0, 0],
        [0, -0.598, 0.801],
        [0, -0.801, -0.598],
    ])
    # test extrinsic direction
    rot_cam_param = rotate_camera(src_cam_param, flip_rot_mat)
    assert rot_cam_param.world2cam == src_cam_param.world2cam
    src_cam_param.inverse_extrinsic()
    rot_inv_cam_param = rotate_camera(src_cam_param, flip_rot_mat)
    assert rot_inv_cam_param.world2cam == src_cam_param.world2cam
    # visualize with projection
    keypoints3d = np.load(
        os.path.join(input_dir, 'keypoints3d.npz'), allow_pickle=True)
    keypoints3d = dict(keypoints3d)
    projector_config = dict(
        mmcv.Config.fromfile('config/ops/projection/opencv_projector.py'))
    projector_config['camera_parameters'] = [
        rot_cam_param,
    ]
    projector = build_projector(projector_config)
    points3d = np.array(keypoints3d['keypoints'][0, 0, :, :3])
    for point_idx, point_loc in enumerate(points3d):
        points3d[point_idx] = np.matmul(flip_rot_mat, point_loc)
    points3d_mask = np.expand_dims(keypoints3d['mask'][0, 0, :], axis=-1)
    projected_points = projector.project(points3d, points_mask=points3d_mask)
    canvas = cv2.imread(os.path.join(input_dir, '000006.png'))
    valid_idxs = np.where(points3d_mask == 1)
    for point_idx in valid_idxs[0]:
        cv2.circle(
            img=canvas,
            center=projected_points[0, point_idx].astype(np.int32),
            radius=2,
            color=(0, 0, 255))
    cv2.imwrite(
        filename=os.path.join(output_dir, 'projected_kps_rotation.jpg'),
        img=canvas)


def test_translate_camera():
    src_cam_param = PinholeCameraParameter.fromfile(
        os.path.join(input_dir, 'cam_006.json'))
    translation = np.array([200, 0, 0], )
    # test extrinsic direction
    transl_cam_param = translate_camera(src_cam_param, translation)
    assert transl_cam_param.world2cam == src_cam_param.world2cam
    src_cam_param.inverse_extrinsic()
    transl_inv_cam_param = translate_camera(src_cam_param, translation)
    assert transl_inv_cam_param.world2cam == src_cam_param.world2cam
    # visualize with projection
    keypoints3d = np.load(
        os.path.join(input_dir, 'keypoints3d.npz'), allow_pickle=True)
    keypoints3d = dict(keypoints3d)
    projector_config = dict(
        mmcv.Config.fromfile('config/ops/projection/opencv_projector.py'))
    projector_config['camera_parameters'] = [
        transl_cam_param,
    ]
    projector = build_projector(projector_config)
    points3d = np.array(keypoints3d['keypoints'][0, 0, :, :3])
    points3d = points3d + np.expand_dims(translation, 0).repeat(
        len(points3d), 0)
    points3d_mask = np.expand_dims(keypoints3d['mask'][0, 0, :], axis=-1)
    projected_points = projector.project(points3d, points_mask=points3d_mask)
    canvas = cv2.imread(os.path.join(input_dir, '000006.png'))
    valid_idxs = np.where(points3d_mask == 1)
    for point_idx in valid_idxs[0]:
        cv2.circle(
            img=canvas,
            center=projected_points[0, point_idx].astype(np.int32),
            radius=2,
            color=(0, 0, 255))
    cv2.imwrite(
        filename=os.path.join(output_dir, 'projected_kps_translation.jpg'),
        img=canvas)



import numpy as np

from xrprimer.data_structure.camera.camera import BaseCameraParameter


def convert_camera_to_opencv(
        cam_param: BaseCameraParameter) -> BaseCameraParameter:
    """Convert a camera parameter instance into opencv convention.

    Args:
        cam_param (BaseCameraParameter):
            The input camera parameter, which is an instance of
            BaseCameraParameter subclass.

    Raises:
        NotImplementedError:
            cam_param.convention has not been supported.

    Returns:
        BaseCameraParameter:
            A camera in the same type as input, whose
            direction is cam2world and convention
            is 'opencv'.
    """
    # do not modify cam_param, modify the cloned cam
    cam_param_backup = cam_param
    cam_param = cam_param_backup.clone()
    if cam_param.world2cam is True:
        cam_param.inverse_extrinsic()
    if cam_param.convention == 'opencv':
        return cam_param
    elif cam_param.convention == 'blender':
        # rotation of euler zxy, (0, 180, 0)
        rot_mat = np.array([[1.0, 0.0, 0.0], [0.0, -1.0, 0.0],
                            [0.0, 0.0, -1.0]])
        extrinsic_r = np.asarray(cam_param.get_extrinsic_r())
        extrinsic_r = np.matmul(rot_mat, extrinsic_r)
        cam_param.set_KRT(R=extrinsic_r)
        cam_param.convention = 'opencv'
        return cam_param
    elif cam_param.convention == 'unreal':
        cam_location = np.asarray(cam_param.get_extrinsic_t())
        # left hand to right hand
        cam_location *= np.asarray((-1, 1, 1))
        # rotation of euler zxy, (90, 270, 0)
        rot_mat = np.array([[0.0, -1.0, 0.0], [0.0, 0.0, 1.0],
                            [-1.0, 0.0, 0.0]])
        extrinsic_r = np.asarray(cam_param.get_extrinsic_r())
        extrinsic_r = np.matmul(rot_mat, extrinsic_r)
        cam_param.set_KRT(R=extrinsic_r, T=cam_location)
        cam_param.convention = 'opencv'
        return cam_param
    else:
        cam_param.logger.error(
            f'Converting a camera from {cam_param.convention} to opencv' +
            ' has not been supported yet.')
        raise NotImplementedError



import numpy as np

from xrprimer.data_structure.camera.camera import BaseCameraParameter


def convert_camera_from_opencv(cam_param: BaseCameraParameter,
                               dst: str) -> BaseCameraParameter:
    """Convert a camera parameter instance from opencv convention into another
    convention.

    Args:
        cam_param (BaseCameraParameter):
            The input camera parameter defined in opencv convention,
            which is an instance of BaseCameraParameter subclass.
        dst (str):
            The name of destination convention.

    Raises:
        NotImplementedError:
            dst convention has not been supported.

    Returns:
        BaseCameraParameter:
            A camera in the same type as input, whose
            direction is cam2world and and convention
            equals to dst.
    """
    assert cam_param.convention == 'opencv'
    # do not modify cam_param, modify the cloned cam
    cam_param_backup = cam_param
    cam_param = cam_param_backup.clone()
    if cam_param.world2cam is True:
        cam_param.inverse_extrinsic()
    if dst == 'opencv':
        return cam_param
    elif dst == 'blender':
        # rotation of euler zxy, (0, 180, 0)
        rot_mat = np.array([[1.0, 0.0, 0.0], [0.0, -1.0, 0.0],
                            [0.0, 0.0, -1.0]])
        extrinsic_r = np.asarray(cam_param.get_extrinsic_r())
        extrinsic_r = np.matmul(rot_mat, extrinsic_r)
        cam_param.set_KRT(R=extrinsic_r)
        cam_param.convention = 'blender'
        return cam_param
    elif dst == 'unreal':
        cam_location = np.asarray(cam_param.get_extrinsic_t())
        # right hand to left hand
        cam_location *= np.asarray((-1, 1, 1))
        # rotation of euler zxy, (270, 0, 90)
        rot_mat = np.array([[0.0, 0.0, 1.0], [-1.0, 0.0, 0.0],
                            [0.0, -1.0, 0.0]])
        extrinsic_r = np.asarray(cam_param.get_extrinsic_r())
        extrinsic_r = np.matmul(rot_mat, extrinsic_r)
        cam_param.set_KRT(R=extrinsic_r, T=cam_location)
        cam_param.convention = 'unreal'
        return cam_param
    else:
        cam_param.logger.error(f'Converting a camera from opencv to {dst}' +
                               ' has not been supported yet.')
        raise NotImplementedError



import os

import numpy as np
import pytest

from xrprimer.data_structure.camera import FisheyeCameraParameter
from xrprimer.transform.convention.camera import (
    convert_camera_parameter,
    downgrade_k_4x4,
    upgrade_k_3x3,
)

cam_param_dir = 'test/data/data_structure/camera_parameter'


def test_intrinsic():
    # test one k
    intrinsic44 = np.asarray([[954.5469360351562, 0.0, 955.3258056640625, 0.0],
                              [0.0, 953.5127563476562, 552.7735595703125, 0.0],
                              [0.0, 0.0, 0.0, 1.0], [0.0, 0.0, 1.0, 0.0]])
    intrinsic33 = downgrade_k_4x4(intrinsic44)
    assert len(intrinsic33) == 3
    assert intrinsic33[2][2] == 1
    intrinsic44_back = upgrade_k_3x3(intrinsic33, is_perspective=True)
    assert np.allclose(intrinsic44_back, intrinsic44)
    # test batched k
    intrinsic44 = np.expand_dims(intrinsic44, axis=0).repeat(5, axis=0)
    intrinsic33 = downgrade_k_4x4(intrinsic44)
    assert len(intrinsic33) == 5
    assert len(intrinsic33[0]) == 3
    assert intrinsic33[0][2][2] == 1
    intrinsic44_back = upgrade_k_3x3(intrinsic33, is_perspective=True)
    assert np.allclose(intrinsic44_back, intrinsic44)


def test_convert_camera():
    fisheye_path = os.path.join(cam_param_dir,
                                'fisheye_param_not_at_origin.json')
    fisheye_opencv_param = FisheyeCameraParameter()
    fisheye_opencv_param.load(fisheye_path)
    # test opencv to opencv
    converted_param = convert_camera_parameter(
        cam_param=fisheye_opencv_param, dst='opencv')
    assert np.allclose(
        np.asarray(converted_param.get_extrinsic_r()),
        np.asarray(fisheye_opencv_param.get_extrinsic_r()))
    # test opencv to blender, world2cam
    if not fisheye_opencv_param.world2cam:
        fisheye_opencv_param.inverse_extrinsic()
    cam_backup = fisheye_opencv_param.clone()
    converted_param = convert_camera_parameter(
        cam_param=fisheye_opencv_param, dst='blender')
    # direction not changed
    assert converted_param.world2cam == fisheye_opencv_param.world2cam
    # input not changed
    assert np.all(
        np.asarray(fisheye_opencv_param.get_extrinsic_r()) == np.asarray(
            cam_backup.get_extrinsic_r()))
    # test opencv to blender, cam2world
    if fisheye_opencv_param.world2cam:
        fisheye_opencv_param.inverse_extrinsic()
    cam_backup = fisheye_opencv_param.clone()
    converted_param = convert_camera_parameter(
        cam_param=fisheye_opencv_param, dst='blender')
    # direction not changed
    assert converted_param.world2cam == fisheye_opencv_param.world2cam
    # input not changed
    assert np.all(
        np.asarray(fisheye_opencv_param.get_extrinsic_r()) == np.asarray(
            cam_backup.get_extrinsic_r()))
    # type not changed
    assert type(fisheye_opencv_param) == type(converted_param)
    with pytest.raises(NotImplementedError):
        converted_param = convert_camera_parameter(
            cam_param=fisheye_opencv_param, dst='ue5')
    with pytest.raises(NotImplementedError):
        converted_param.convention = 'ue5'
        converted_param = convert_camera_parameter(
            cam_param=converted_param, dst='opencv')



import os
import shutil

import cv2
import numpy as np
import pytest

from xrprimer.utils.ffmpeg_utils import (
    VideoInfoReader,
    VideoWriter,
    array_to_images,
    array_to_video,
    images_to_array,
    images_to_array_opencv,
    pad_for_libx264,
    video_to_array,
)

input_dir = 'test/data/utils/test_ffmpeg_utils'
output_dir = 'test/data/output/utils/test_ffmpeg_utils'


@pytest.fixture(scope='module', autouse=True)
def fixture():
    if os.path.exists(output_dir):
        shutil.rmtree(output_dir)
    os.makedirs(output_dir, exist_ok=False)
    test_video_writter = VideoWriter(
        os.path.join(output_dir, 'test_video.mp4'),
        resolution=[256, 512],
        fps=15)
    for _ in range(75):
        test_video_writter.write(
            np.random.randint(
                low=0, high=255, size=(256, 512, 3), dtype=np.uint8))
    test_video_writter.close()
    test_frames_dir = os.path.join(output_dir, 'test_frames')
    os.mkdir(test_frames_dir)
    for i in range(25):
        test_img = np.random.randint(
            low=0, high=255, size=(256, 512, 3), dtype=np.uint8)
        cv2.imwrite(
            filename=os.path.join(test_frames_dir, f'{i:06d}.png'),
            img=test_img)


def test_pad_for_libx264():
    gray_image = np.ones(shape=[25, 45], dtype=np.uint8)
    pad_gray_image = pad_for_libx264(gray_image)
    assert pad_gray_image.shape[0] % 2 == 0
    assert pad_gray_image.shape[1] % 2 == 0

    rgb_image = np.ones(shape=[25, 45, 3], dtype=np.uint8)
    pad_rgb_image = pad_for_libx264(rgb_image)
    assert pad_rgb_image.shape[0] % 2 == 0
    assert pad_rgb_image.shape[1] % 2 == 0

    gray_array = np.ones(shape=[11, 25, 45], dtype=np.uint8)
    pad_gray_array = pad_for_libx264(gray_array)
    assert pad_gray_array.shape[1] % 2 == 0
    assert pad_gray_array.shape[2] % 2 == 0

    rgb_array = np.ones(shape=[11, 25, 45, 3], dtype=np.uint8)
    pad_rgb_array = pad_for_libx264(rgb_array)
    assert pad_rgb_array.shape[1] % 2 == 0


def test_video_to_array():
    test_video_path = os.path.join(output_dir, 'test_video.mp4')
    # test images_to_array
    img_arr = video_to_array(test_video_path)
    assert img_arr.shape == (75, 256, 512, 3)
    img_arr = video_to_array(test_video_path, resolution=(128, 128))
    assert img_arr.shape == (75, 128, 128, 3)
    img_arr = video_to_array(test_video_path, start=0, end=20)
    assert img_arr.shape == (20, 256, 512, 3)


def test_images_to_array():
    test_frames_dir = os.path.join(output_dir, 'test_frames')
    # test images_to_array
    img_arr = images_to_array(test_frames_dir)
    assert img_arr.shape == (25, 256, 512, 3)
    img_arr = images_to_array(test_frames_dir, resolution=(128, 128))
    assert img_arr.shape == (25, 128, 128, 3)
    img_arr = images_to_array(
        test_frames_dir, start=0, end=20, img_format=None)
    assert img_arr.shape == (20, 256, 512, 3)
    # test images_to_array_opencv
    img_arr = images_to_array_opencv(test_frames_dir)
    assert img_arr.shape == (25, 256, 512, 3)
    img_arr = images_to_array_opencv(test_frames_dir, resolution=(128, 128))
    assert img_arr.shape == (25, 128, 128, 3)
    img_arr = images_to_array_opencv(test_frames_dir, start=0, end=20)
    assert img_arr.shape == (20, 256, 512, 3)
    img_arr = images_to_array_opencv(test_frames_dir, start=0, end=20)
    assert img_arr.shape == (20, 256, 512, 3)


def test_array_to_images():
    img_arr = np.random.randint(
        low=0, high=255, size=(3, 256, 512, 3), dtype=np.uint8)
    # test number of images
    path = os.path.join(output_dir, 'test_array_to_images')
    array_to_images(img_arr, output_folder=path)
    assert len(os.listdir(path)) == 3
    # test img_format
    path = os.path.join(output_dir, 'test_array_to_images_format')
    array_to_images(img_arr, output_folder=path, img_format='frame_%06d.jpg')
    assert sorted(os.listdir(path))[0] == 'frame_000000.jpg'
    # test resolution
    with pytest.raises(NotImplementedError):
        path = os.path.join(output_dir, 'test_array_to_images_resolution')
        array_to_images(img_arr, output_folder=path, resolution=(512, 1024))


def test_array_to_video():
    img_arr = np.random.randint(
        low=0, high=255, size=(3, 256, 512, 3), dtype=np.uint8)
    # test number of frames
    path = os.path.join(output_dir, 'test_array_to_video.mp4')
    array_to_video(img_arr, output_path=path)
    reader = VideoInfoReader(path)
    assert int(reader['nb_frames']) == 3
    assert int(reader['height']) == 256
    assert int(reader['width']) == 512
    # test resolution
    path = os.path.join(output_dir, 'test_array_to_video_resolution.mp4')
    array_to_video(img_arr, output_path=path, resolution=(128, 256))
    reader = VideoInfoReader(path)
    assert int(reader['height']) == 128
    assert int(reader['width']) == 256
    # test fps
    path = os.path.join(output_dir, 'test_array_to_video_fps.mp4')
    array_to_video(img_arr, output_path=path, fps=10)
    reader = VideoInfoReader(path)
    assert int(reader['nb_frames']) == 3
    assert int(reader['avg_frame_rate'].split('/')[0]) == 10



import os
import shutil

import cv2
import numpy as np
import pytest

from xrprimer.data_structure.camera import (
    FisheyeCameraParameter,
    OmniCameraParameter,
    PinholeCameraParameter,
)
from xrprimer.transform.camera.distortion import (
    undistort_camera,
    undistort_images,
    undistort_points,
)

input_dir = 'test/data/transform/camera/test_distortion'
output_dir = 'test/data/output/transform/camera/test_distortion'


@pytest.fixture(scope='module', autouse=True)
def fixture():
    if os.path.exists(output_dir):
        shutil.rmtree(output_dir)
    os.makedirs(output_dir, exist_ok=False)


def test_undistort_camera():
    # test only distort camera parameter
    input_json_path = os.path.join(input_dir, 'dist_fisheye_param.json')
    fisheye_param = FisheyeCameraParameter(name='distort')
    fisheye_param.load(input_json_path)
    fisheye_param.set_KRT(R=np.eye(3))
    intrinsic_backup = np.asarray(fisheye_param.get_intrinsic())
    new_cam_param = undistort_camera(distorted_cam=fisheye_param)
    assert isinstance(new_cam_param, PinholeCameraParameter)
    # test input not changed
    assert np.all(
        intrinsic_backup == np.asarray(fisheye_param.get_intrinsic()))
    # test raises
    pytorch3d_cam = fisheye_param.clone()
    pytorch3d_cam.convention = 'pytorch3d'
    with pytest.raises(NotImplementedError):
        new_cam_param, new_img = undistort_camera(distorted_cam=pytorch3d_cam)
    omni_cam = OmniCameraParameter(convention='opencv')
    with pytest.raises(AssertionError):
        new_cam_param = undistort_camera(distorted_cam=omni_cam)


def test_undistort_images():
    fisheye_param = FisheyeCameraParameter(name='distort')
    fisheye_param.load(os.path.join(input_dir, 'dist_fisheye_param.json'))
    fisheye_param.set_KRT(R=np.eye(3))
    intrinsic_backup = np.asarray(fisheye_param.get_intrinsic())
    test_img = cv2.imread(filename=os.path.join(input_dir, 'dist_img.png'))
    test_imgs = np.expand_dims(test_img, axis=0)
    test_imgs = np.repeat(test_imgs, 2, axis=0)
    img_backup = test_img.copy()
    new_cam_param, new_img = undistort_images(
        distorted_cam=fisheye_param, image_array=test_imgs)
    # test input not changed
    assert np.all(
        intrinsic_backup == np.asarray(fisheye_param.get_intrinsic()))
    assert np.all(img_backup == test_imgs)
    # test output
    assert isinstance(new_cam_param, PinholeCameraParameter)
    assert np.all(new_img.shape == test_imgs.shape)
    undist_img = new_img[0]
    cv2.imwrite(
        filename=os.path.join(output_dir, 'undist_img.jpg'), img=undist_img)
    # test raises
    pytorch3d_cam = fisheye_param.clone()
    pytorch3d_cam.convention = 'pytorch3d'
    with pytest.raises(NotImplementedError):
        new_cam_param, new_img = undistort_images(
            distorted_cam=pytorch3d_cam, image_array=test_imgs)


def test_undistort_points():
    fisheye_param = FisheyeCameraParameter(name='distort')
    fisheye_param.load(os.path.join(input_dir, 'dist_fisheye_param.json'))
    fisheye_param.set_KRT(R=np.eye(3))
    intrinsic_backup = np.asarray(fisheye_param.get_intrinsic())
    corners_points = np.array([[791, 78], [1196, 19], [1215, 494], [1167, 534],
                               [851, 529], [808, 491]],
                              dtype=np.int32)
    points_backup = corners_points.copy()
    new_cam_param, new_points = undistort_points(
        distorted_cam=fisheye_param, points=corners_points)
    # test input not changed
    assert np.all(
        intrinsic_backup == np.asarray(fisheye_param.get_intrinsic()))
    assert np.all(points_backup == corners_points)
    # test output
    assert isinstance(new_cam_param, PinholeCameraParameter)
    assert np.all(new_points.shape == corners_points.shape)
    # visualize results
    distort_img = cv2.imread(filename=os.path.join(input_dir, 'dist_img.png'))
    _, new_img = undistort_images(
        distorted_cam=fisheye_param,
        image_array=np.expand_dims(distort_img, axis=0))
    for point in corners_points.astype(np.int32):
        cv2.circle(img=distort_img, center=point, radius=3, color=(0, 0, 255))
    cv2.imwrite(
        filename=os.path.join(output_dir, 'dist_img_with_points.jpg'),
        img=distort_img)
    new_img = new_img[0]
    for point in new_points.astype(np.int32):
        cv2.circle(img=new_img, center=point, radius=3, color=(0, 0, 255))
    cv2.imwrite(
        filename=os.path.join(output_dir, 'undist_img_with_points.jpg'),
        img=new_img)
    # test another shape
    test_points = np.ones(shape=[4, 3, 2, 2])
    points_backup = test_points.copy()
    new_cam_param, new_points = undistort_points(
        distorted_cam=fisheye_param, points=test_points)
    # test input not changed
    assert np.all(
        intrinsic_backup == np.asarray(fisheye_param.get_intrinsic()))
    assert np.all(points_backup == test_points)
    # test output
    assert isinstance(new_cam_param, PinholeCameraParameter)
    assert np.all(new_points.shape == test_points.shape)
    # test raises
    pytorch3d_cam = fisheye_param.clone()
    pytorch3d_cam.convention = 'pytorch3d'
    with pytest.raises(NotImplementedError):
        new_cam_param, new_points = undistort_points(
            distorted_cam=pytorch3d_cam, points=test_points)



import logging
from typing import Union

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')


def setup_logger(logger_name: str = 'root',
                 logger_level: int = logging.INFO,
                 logger_path: str = None,
                 logger_format: str = None) -> logging.Logger:
    """Set up a logger.

    Args:
        logger_name (str, optional):
            Name of the logger. Defaults to 'root'.
        logger_level (int, optional):
            Set the logging level of this logger.
            Defaults to logging.INFO.
        logger_path (str, optional):
            Path to the log file.
            Defaults to None, no file will be written,
            StreamHandler will be used.
        logger_format (str, optional):
            The formatter for logger handler.
            Defaults to None.

    Returns:
        logging.Logger:
            A logger with settings above.
    """
    logger = logging.getLogger(logger_name)
    logger.setLevel(level=logger_level)
    # prevent logging twice in stdout
    logger.propagate = False
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.INFO)
    handlers = [stream_handler]
    if logger_path is not None:
        handler = logging.FileHandler(logger_path)
        handlers.append(handler)
    if logger_format is not None:
        formatter = logging.Formatter(logger_format)
    else:
        formatter = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    # assure handlers are not double
    while logger.hasHandlers():
        logger.removeHandler(logger.handlers[0])
    for handler in handlers:
        handler.setFormatter(formatter)
        logger.addHandler(handler)
    return logger


def get_logger(
        logger: Union[None, str, logging.Logger] = None) -> logging.Logger:
    """Get logger.

    Args:
        logger (Union[None, str, logging.Logger]):
            None for root logger. Besides, pass name of the
            logger or the logger itself.
            Defaults to None.

    Returns:
        logging.Logger
    """
    if logger is None or isinstance(logger, str):
        ret_logger = logging.getLogger(logger)
    else:
        ret_logger = logger
    return ret_logger



# yapf: disable
import glob
import json
import logging
import os
import shutil
import subprocess
from pathlib import Path
from typing import Any, Tuple, Union

import cv2
import numpy as np

from .log_utils import get_logger
from .path_utils import (
    Existence,
    check_path,
    check_path_suffix,
    prepare_output_path,
)

# yapf: enable


class VideoInfoReader():
    INFO_KEYS = [
        'index', 'codec_name', 'codec_long_name', 'profile', 'codec_type',
        'codec_time_base', 'codec_tag_string', 'codec_tag', 'width', 'height',
        'coded_width', 'coded_height', 'has_b_frames', 'pix_fmt', 'level',
        'chroma_location', 'refs', 'is_avc', 'nal_length_size', 'r_frame_rate',
        'avg_frame_rate', 'time_base', 'start_pts', 'start_time',
        'duration_ts', 'duration', 'bit_rate', 'bits_per_raw_sample',
        'nb_frames', 'disposition', 'tags'
    ]

    def __init__(self,
                 input_path: str,
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Get video information from video, mimiced from ffmpeg-python.
        https://github.com/kkroening/ffmpeg-python.

        Args:
            vid_file (str): Path to the video file.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.

        Raises:
            FileNotFoundError: check the input path.

        Returns:
            None.
        """
        self.logger = get_logger(logger)
        check_path(
            input_path=input_path,
            allowed_existence=[Existence.FileExist],
            allowed_suffix=[
                '.mp4', '.mkv', '.avi', '.gif', '.png', '.jpg', '.jpeg'
            ],
            path_type='file',
            logger=logger)
        cmd = [
            'ffprobe', '-show_format', '-show_streams', '-of', 'json',
            input_path
        ]
        process = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, _ = process.communicate()
        probe = json.loads(out.decode('utf-8'))
        video_stream = next((stream for stream in probe['streams']
                             if stream['codec_type'] == 'video'), None)
        if video_stream is None:
            self.logger.error('No video stream found')
            raise ValueError
        self.video_stream = video_stream

    def __getitem__(self, key: str) -> Any:
        """Get the corresponding information according to the key.

        Args:
            key (str):
                A key in VideoInfoReader.INFO_KEYS
                Such as codec_name, pix_fmt, duration, etc.

        Raises:
            KeyError: key cannot be found in VideoInfoReader.INFO_KEYS

        Returns:
            Any: The expected information.
        """
        if key not in self.__class__.INFO_KEYS:
            self.logger.error(
                'Wronge vid info key.' +
                f'Select one key from {self.__class__.INFO_KEYS}')
            raise KeyError
        return self.video_stream[key]


class VideoWriter:

    def __init__(self,
                 output_path: str,
                 resolution: Tuple[int, int],
                 fps: float = 30.0,
                 n_frames: int = 1e9,
                 disable_log: bool = False,
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Write video file by ffmpeg.

        Args:
            output_path (str):
                Path to the output video file, which shall end with '.mp4'.
            resolution (List[int, int]):
                Resolution of the video, [height, width].
            fps (float, optional):
                Frame per second. Defaults to 30.0.
            n_frames (int, optional):
                Limit number of frames in this video.
                Defaults to 1e9.
            disable_log (bool, optional):
                Whether to disable logs of info level.
                Defaults to False.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.

        Raises:
            BrokenPipeError: No buffer received.
        """
        self.logger = get_logger(logger)
        prepare_output_path(
            output_path,
            allowed_suffix=['.mp4'],
            tag='output video',
            path_type='file',
            overwrite=True)
        height, width = resolution
        width += width % 2
        height += height % 2
        command = [
            'ffmpeg',
            '-y',  # (optional) overwrite output file if it exists
            '-f',
            'rawvideo',
            '-pix_fmt',
            'bgr24',
            '-s',
            f'{int(width)}x{int(height)}',
            '-r',
            f'{fps}',  # frames per second
            '-loglevel',
            'error',
            '-threads',
            '1',
            '-i',
            '-',  # The input comes from a pipe
            '-vcodec',
            'libx264',
            '-r',
            f'{fps}',  # frames per second
            '-an',  # Tells FFMPEG not to expect any audio
            output_path,
        ]
        if not disable_log:
            self.logger.info(f'Running \"{" ".join(command)}\"')
        process = subprocess.Popen(
            command,
            stdin=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        if process.stdin is None or process.stderr is None:
            self.logger.error('No buffer received.')
            raise BrokenPipeError
        self.process = process
        self.n_frames = n_frames
        self.len = 0

    def write(self, image_array: np.ndarray):
        """Write an image to the video file.

        Args:
            image_array (np.ndarray):
                An array of a single image, in shape
                [H, W, C].
        """
        if self.len < self.n_frames:
            try:
                self.process.stdin.write(image_array.tobytes())
                self.len += 1
            except KeyboardInterrupt:
                self.__del__()
        if self.len >= self.n_frames:
            self.__del__()

    def __del__(self):
        self.process.stdin.close()
        self.process.stderr.close()
        self.process.wait()

    def close(self):
        """Manually close this video writer."""
        self.__del__()


def pad_for_libx264(image_array: np.ndarray) -> np.ndarray:
    """Pad zeros if width or height of image_array is not divisible by 2.
    Otherwise you will get.

    \"[libx264 @ 0x1b1d560] width not divisible by 2 \"

    Args:
        image_array (np.ndarray):
            Image or images load by cv2.imread().
            Possible shapes:
            1. [height, width]
            2. [height, width, channels]
            3. [images, height, width]
            4. [images, height, width, channels]

    Returns:
        np.ndarray:
            A image with both edges divisible by 2.
    """
    if image_array.ndim == 2 or \
            (image_array.ndim == 3 and image_array.shape[2] == 3):
        hei_index = 0
        wid_index = 1
    elif image_array.ndim == 4 or \
            (image_array.ndim == 3 and image_array.shape[2] != 3):
        hei_index = 1
        wid_index = 2
    else:
        return image_array
    hei_pad = image_array.shape[hei_index] % 2
    wid_pad = image_array.shape[wid_index] % 2
    if hei_pad + wid_pad > 0:
        pad_width = []
        for dim_index in range(image_array.ndim):
            if dim_index == hei_index:
                pad_width.append((0, hei_pad))
            elif dim_index == wid_index:
                pad_width.append((0, wid_pad))
            else:
                pad_width.append((0, 0))
        values = 0
        image_array = \
            np.pad(image_array,
                   pad_width,
                   mode='constant', constant_values=values)
    return image_array


def video_to_array(
        input_path: str,
        resolution: Union[Tuple[int, int], Tuple[float, float]] = None,
        start: int = 0,
        end: int = None,
        disable_log: bool = False,
        logger: Union[None, str, logging.Logger] = None) -> np.ndarray:
    """
    Read a video/gif as an array of (f * h * w * 3).

    Args:
        input_path (str): input path.
        resolution (Union[Tuple[int, int], Tuple[float, float]],
            optional): resolution(height, width) of output.
            Defaults to None.
        start (int, optional): start frame index. Inclusive.
             If < 0, will be converted to frame_index range in [0, n_frame].
            Defaults to 0.
        end (int, optional): end frame index. Exclusive.
            Could be positive int or negative int or None.
            If None, all frames from start till the last frame are included.
            Defaults to None.
        disable_log (bool, optional): whether close the ffmepg command info.
            Defaults to False.
        logger (Union[None, str, logging.Logger], optional):
            Logger for logging. If None, root logger will be selected.
            Defaults to None.

    Raises:
        FileNotFoundError: check the input path.

    Returns:
        np.ndarray: shape will be (f * h * w * 3).
    """
    info = VideoInfoReader(input_path, logger=logger)
    if resolution:
        height, width = resolution
    else:
        width, height = int(info['width']), int(info['height'])
    n_frames = int(info['nb_frames'])
    start = (min(start, n_frames - 1) + n_frames) % n_frames
    end = (min(end, n_frames - 1) +
           n_frames) % n_frames if end is not None else n_frames
    command = [
        'ffmpeg',
        '-i',
        input_path,
        '-filter_complex',
        f'[0]trim=start_frame={start}:end_frame={end}[v0]',
        '-map',
        '[v0]',
        '-pix_fmt',
        'bgr24',  # bgr24 for matching OpenCV
        '-s',
        f'{int(width)}x{int(height)}',
        '-f',
        'image2pipe',
        '-vcodec',
        'rawvideo',
        '-loglevel',
        'error',
        'pipe:'
    ]
    if not disable_log:
        logger = get_logger(logger)
        logger.info(f'Running \"{" ".join(command)}\"')
    # Execute FFmpeg as sub-process with stdout as a pipe
    process = subprocess.Popen(command, stdout=subprocess.PIPE, bufsize=10**8)
    if process.stdout is None:
        raise BrokenPipeError('No buffer received.')
    # Read decoded video frames from the PIPE until no more frames to read
    array = []
    while True:
        # Read decoded video frame (in raw video format) from stdout process.
        buffer = process.stdout.read(int(width * height * 3))
        # Break the loop if buffer length is not W*H*3\
        # (when FFmpeg streaming ends).
        if len(buffer) != width * height * 3:
            break
        img = np.frombuffer(buffer, np.uint8).reshape(height, width, 3)
        array.append(img[np.newaxis])
    process.stdout.flush()
    process.stdout.close()
    process.wait()
    return np.concatenate(array)


def images_to_array_opencv(
        input_folder: str,
        resolution: Union[Tuple[int, int], Tuple[float, float]] = None,
        img_format: Union[str, None] = None,
        start: int = 0,
        end: int = None,
        logger: Union[None, str, logging.Logger] = None) -> np.ndarray:
    """
    Read a folder of images as an array of (f * h * w * 3).

    Args:
        input_folder (str): folder of input images.
        resolution (Union[Tuple[int, int], Tuple[float, float]]):
            resolution(height, width) of output. Defaults to None.
        img_format (str, optional):
            Format of images to be read, 'jpg' or 'png'.
            Defaults to None.
        start (int, optional): start frame index. Inclusive.
             If < 0, will be converted to frame_index range in [0, n_frame].
            Defaults to 0.
        end (int, optional): end frame index. Exclusive.
            Could be positive int or negative int or None.
            If None, all frames from start till the last frame are included.
            Defaults to None.
        logger (Union[None, str, logging.Logger], optional):
            Logger for logging. If None, root logger will be selected.
            Defaults to None.
    Raises:
        FileNotFoundError: check the input path.

    Returns:
        np.ndarray: shape will be (f * h * w * 3).
    """
    check_path(
        input_path=input_folder,
        allowed_existence=[Existence.DirectoryExistNotEmpty],
        allowed_suffix=[''],
        path_type='dir',
        logger=logger)
    if img_format is None:
        frame_list = []
        frame_names = sorted(os.listdir(input_folder))
        for name in frame_names:
            abs_path = os.path.join(input_folder, name)
            if check_path_suffix(abs_path, ['.jpg', '.jpeg', '.png']):
                frame_list.append(abs_path)
    else:
        frame_list = sorted(
            glob.glob(os.path.join(input_folder, f'*.{img_format}')))
    if end is None:
        frame_list = frame_list[start:]
    else:
        frame_list = frame_list[start:end]
    array_list = []
    for index, frame_path in enumerate(frame_list):
        img = cv2.imread(frame_path)
        if index == 0 and resolution is None:
            resolution = img.shape[0:2]
        else:
            img = cv2.resize(img, (resolution[1], resolution[0]))
        array_list.append(img)
    return np.asarray(array_list)


def images_to_array(
        input_folder: str,
        resolution: Union[Tuple[int, int], Tuple[float, float]] = None,
        img_format: str = '%06d.png',
        start: int = 0,
        end: int = None,
        remove_raw_files: bool = False,
        disable_log: bool = False,
        logger: Union[None, str, logging.Logger] = None) -> np.ndarray:
    """
    Read a folder of images as an array of (f * h * w * 3).

    Args:
        input_folder (str): folder of input images.
        resolution (Union[Tuple[int, int], Tuple[float, float]]):
            resolution(height, width) of output. Defaults to None.
        img_format (str, optional): format of images to be read.
            Defaults to '%06d.png'.
        start (int, optional): start frame index. Inclusive.
             If < 0, will be converted to frame_index range in [0, n_frame].
            Defaults to 0.
        end (int, optional): end frame index. Exclusive.
            Could be positive int or negative int or None.
            If None, all frames from start till the last frame are included.
            Defaults to None.
        remove_raw_files (bool, optional): whether remove raw images.
            Defaults to False.
        disable_log (bool, optional): whether close the ffmepg command info.
            Defaults to False.
    Raises:
        FileNotFoundError: check the input path.

    Returns:
        np.ndarray: shape will be (f * h * w * 3).
    """
    check_path(
        input_path=input_folder,
        allowed_existence=[Existence.DirectoryExistNotEmpty],
        allowed_suffix=[''],
        path_type='dir',
        logger=logger)

    input_folder_info = Path(input_folder)

    temp_input_folder = None
    if img_format is None:
        temp_input_folder = os.path.join(input_folder_info.parent,
                                         input_folder_info.name + '_temp')
        img_format = images_to_sorted_images(
            input_folder=input_folder, output_folder=temp_input_folder)
        input_folder = temp_input_folder

    info = VideoInfoReader(f'{input_folder}/{img_format}' % start)
    width, height = int(info['width']), int(info['height'])
    if resolution:
        height, width = resolution
    else:
        width, height = int(info['width']), int(info['height'])

    n_frames = len(os.listdir(input_folder))
    start = max(start, 0) % n_frames
    end = min(end, n_frames) % (n_frames + 1) \
        if end is not None else n_frames
    command = [
        'ffmpeg',
        '-y',
        '-threads',
        '1',
        '-start_number',
        f'{start}',
        '-i',
        f'{input_folder}/{img_format}',
        '-frames:v',
        f'{end - start}',
        '-f',
        'rawvideo',
        '-pix_fmt',
        'bgr24',  # bgr24 for matching OpenCV
        '-s',
        f'{int(width)}x{int(height)}',
        '-loglevel',
        'error',
        '-'
    ]
    if not disable_log:
        logger = get_logger(logger)
        logger.info(f'Running \"{" ".join(command)}\"')
    process = subprocess.Popen(command, stdout=subprocess.PIPE, bufsize=10**8)
    if process.stdout is None:
        raise BrokenPipeError('No buffer received.')
    # Read decoded video frames from the PIPE until no more frames to read
    array = []
    while True:
        # Read decoded video frame (in raw video format) from stdout process.
        buffer = process.stdout.read(int(width * height * 3))
        # Break the loop if buffer length is not W*H*3\
        # (when FFmpeg streaming ends).

        if len(buffer) != width * height * 3:
            break
        img = np.frombuffer(buffer, np.uint8).reshape(height, width, 3)
        array.append(img[np.newaxis])
    process.stdout.flush()
    process.stdout.close()
    process.wait()
    if temp_input_folder is not None and\
            os.path.isdir(temp_input_folder):
        shutil.rmtree(temp_input_folder)
    if remove_raw_files and\
            os.path.isdir(input_folder):
        shutil.rmtree(input_folder)

    return np.concatenate(array)


def images_to_sorted_images(input_folder, output_folder, img_format='%06d'):
    """Copy and rename a folder of images into a new folder following the
    `img_format`.

    Args:
        input_folder (str): input folder.
        output_folder (str): output folder.
        img_format (str, optional): image format name, do not need extension.
            Defaults to '%06d'.

    Returns:
        str: image format of the rename images.
    """
    img_format = img_format.rsplit('.', 1)[0]
    file_list = []
    os.makedirs(output_folder, exist_ok=True)
    pngs = glob.glob(os.path.join(input_folder, '*.png'))
    if pngs:
        ext = 'png'
    file_list.extend(pngs)
    jpgs = glob.glob(os.path.join(input_folder, '*.jpg'))
    if jpgs:
        ext = 'jpg'
    file_list.extend(jpgs)
    file_list.sort()
    for index, file_name in enumerate(file_list):
        shutil.copy(
            file_name,
            os.path.join(output_folder, (img_format + '.%s') % (index, ext)))
    return img_format + '.%s' % ext


def array_to_video(image_array: np.ndarray,
                   output_path: str,
                   fps: Union[int, float] = 30,
                   resolution: Union[Tuple[int, int], Tuple[float,
                                                            float]] = None,
                   disable_log: bool = False,
                   logger: Union[None, str, logging.Logger] = None) -> None:
    """Convert an array to a video directly, gif not supported.

    Args:
        image_array (np.ndarray): shape should be (f * h * w * 3).
        output_path (str): output video file path.
        fps (Union[int, float, optional): fps. Defaults to 30.
        resolution (Optional[Union[Tuple[int, int], Tuple[float, float]]],
            optional): (height, width) of the output video.
            Defaults to None.
        disable_log (bool, optional): whether close the ffmepg command info.
            Defaults to False.
    Raises:
        FileNotFoundError: check output path.
        TypeError: check input array.

    Returns:
        None.
    """
    if not isinstance(image_array, np.ndarray):
        raise TypeError('Input should be np.ndarray.')
    assert image_array.ndim == 4
    assert image_array.shape[-1] == 3
    prepare_output_path(
        output_path,
        allowed_suffix=['.mp4'],
        tag='output video',
        path_type='file',
        overwrite=True,
        logger=logger)
    if resolution:
        height, width = resolution
        width += width % 2
        height += height % 2
    else:
        image_array = pad_for_libx264(image_array)
        height, width = image_array.shape[1], image_array.shape[2]
    command = [
        'ffmpeg',
        '-y',  # (optional) overwrite output file if it exists
        '-f',
        'rawvideo',
        '-s',
        f'{int(width)}x{int(height)}',  # size of one frame
        '-pix_fmt',
        'bgr24',
        '-r',
        f'{fps}',  # frames per second
        '-loglevel',
        'error',
        '-threads',
        '4',
        '-i',
        '-',  # The input comes from a pipe
        '-vcodec',
        'libx264',
        '-an',  # Tells FFMPEG not to expect any audio
        output_path,
    ]
    if not disable_log:
        logger = get_logger(logger)
        logger.info(f'Running \"{" ".join(command)}\"')
    process = subprocess.Popen(
        command,
        stdin=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    if process.stdin is None or process.stderr is None:
        raise BrokenPipeError('No buffer received.')
    index = 0
    while True:
        if index >= image_array.shape[0]:
            break
        process.stdin.write(image_array[index].tobytes())
        index += 1
    process.stdin.close()
    process.stderr.close()
    process.wait()


def array_to_images(image_array: np.ndarray,
                    output_folder: str,
                    img_format: str = '%06d.png',
                    resolution: Union[Tuple[int, int], Tuple[float,
                                                             float]] = None,
                    disable_log: bool = False,
                    logger: Union[None, str, logging.Logger] = None) -> None:
    """Convert an array to images directly.

    Args:
        image_array (np.ndarray): shape should be (f * h * w * 3).
        output_folder (str): output folder for the images.
        img_format (str, optional): format of the images.
            Defaults to '%06d.png'.
        resolution (Optional[Union[Tuple[int, int], Tuple[float, float]]],
            optional): resolution(height, width) of output.
            Defaults to None.
        disable_log (bool, optional): whether close the ffmepg command info.
            Defaults to False.

    Raises:
        FileNotFoundError: check output folder.
        TypeError: check input array.

    Returns:
        None
    """
    logger = get_logger(logger)
    prepare_output_path(
        output_folder,
        allowed_suffix=[],
        tag='output image folder',
        path_type='dir',
        overwrite=True)

    if not isinstance(image_array, np.ndarray):
        raise TypeError('Input should be np.ndarray.')
    assert image_array.ndim == 4
    assert image_array.shape[-1] == 3
    if resolution:
        height, width = resolution
        logger.error('Resolution not correctly implemented.')
        raise NotImplementedError
    else:
        height, width = image_array.shape[1], image_array.shape[2]
    command = [
        'ffmpeg',
        '-y',  # (optional) overwrite output file if it exists
        '-f',
        'rawvideo',
        '-s',
        f'{int(width)}x{int(height)}',  # size of one frame
        '-pix_fmt',
        'bgr24',  # bgr24 for matching OpenCV
        '-loglevel',
        'error',
        '-threads',
        '4',
        '-i',
        '-',  # The input comes from a pipe
        '-f',
        'image2',
        '-start_number',
        '0',
        os.path.join(output_folder, img_format),
    ]
    if not disable_log:
        logger = get_logger(logger)
        logger.info(f'Running \"{" ".join(command)}\"')
    process = subprocess.Popen(
        command,
        stdin=subprocess.PIPE,
        stderr=subprocess.PIPE,
        bufsize=10**8,
        close_fds=True)
    if process.stdin is None or process.stderr is None:
        raise BrokenPipeError('No buffer received.')
    index = 0
    while True:
        if index >= image_array.shape[0]:
            break
        process.stdin.write(image_array[index].tobytes())
        index += 1
    process.stdin.close()
    process.stderr.close()
    process.wait()



import copy
import logging
import os
from typing import Union

import numpy as np

from xrprimer.utils.log_utils import get_logger
from xrprimer_cpp.camera import BaseCameraParameter as BaseCameraParameter_cpp


class BaseCameraParameter(BaseCameraParameter_cpp):
    ATTR_NAMES = [
        'name', 'intrinsic', 'extrinsic_r', 'extrinsic_t', 'height', 'width',
        'world2cam', 'convention'
    ]

    def __init__(self,
                 K: Union[list, np.ndarray, None] = None,
                 R: Union[list, np.ndarray, None] = None,
                 T: Union[list, np.ndarray, None] = None,
                 name: str = 'default',
                 height: int = 1080,
                 width: int = 1920,
                 world2cam: bool = True,
                 convention: str = 'opencv',
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Base class for all python camera parameter classes. Common methods
        are defined in this class.

        Args:
            K (Union[list, np.ndarray, None], optional):
                Nested list of float32, 4x4 or 3x3 K mat.
                Defaults to None, 4x4 zeros.
            R (Union[list, np.ndarray, None], optional):
                Nested list of float32, 3x3 rotation mat.
                Defaults to None, 3x3 identity.
            T (Union[list, np.ndarray, None], optional):
                List of float32, T vector.
                Defaults to None, zero vector.
            name (str, optional):
                Name of this camera. Defaults to 'default'.
            height (int, optional):
                Height of the image shot by this camera.
                Defaults to 1080.
            width (int, optional):
                Width of the image shot by this camera.
                Defaults to 1920.
            world2cam (bool, optional):
                Whether the R, T transform points from world space
                to camera space. Defaults to True.
            convention (str, optional):
                Convention name of this camera.
                Defaults to 'opencv'.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        super().__init__(
            intrinsic=np.zeros((4, 4)),
            extrinsic_r=np.zeros((3, 3)),
            extrinsic_t=np.ones((3, )))
        self.name = name
        self.set_KRT(K, R, T)
        self.set_resolution(height=height, width=width)
        self.world2cam = world2cam
        self.convention = convention
        self.logger = get_logger(logger)

    def set_intrinsic(self,
                      mat3x3: Union[list, np.ndarray, None] = None,
                      width: int = None,
                      height: int = None,
                      fx: float = None,
                      fy: float = None,
                      cx: float = None,
                      cy: float = None,
                      perspective: bool = True) -> None:
        """Set the intrinsic of a camera. Note that mat3x3 has a higher
        priority than fx, fy, cx, cy.

        Args:
            mat3x3 (list, optional):
                A nested list of intrinsic matrix,
                in shape (3, 3). If mat is given,
                fx, fy, cx, cy will be ignored.
                Defaults to None.
            width (int):
                Width of the screen.
            height (int):
                Height of the screen.
            fx (float, optional):
                Focal length. Defaults to None.
            fy (float, optional):
                Focal length. Defaults to None.
            cx (float, optional):
                Camera principal point. Defaults to None.
            cy (float, optional):
                Camera principal point. Defaults to None.
            perspective (bool, optional):
                Whether it is a perspective camera, if not,
                it's orthographics. Defaults to True.
        """
        if mat3x3 is not None:
            mat3x3 = np.asarray(mat3x3)
            super().set_intrinsic(mat3x3=mat3x3, perspective=perspective)
        elif width is not None and\
                height is not None and\
                fx is not None and\
                fy is not None and\
                cx is not None and\
                cy is not None:
            super().set_intrinsic(
                width=width, height=height, fx=fx, fy=fy, cx=cx, cy=cy)
        else:
            self.logger.error(
                'Either mat3x3 or (h, w, fx/y, cx/y) should be offered.')
            raise ValueError

    def set_resolution(self, height: int, width: int) -> None:
        """Set resolution of the camera.

        Args:
            height (int):
                Height of the screen.
            width (int):
                Width of the screen.
        """
        self.height = height
        self.width = width

    def set_KRT(self,
                K: Union[list, np.ndarray, None] = None,
                R: Union[list, np.ndarray, None] = None,
                T: Union[list, np.ndarray, None] = None,
                world2cam: Union[bool, None] = None) -> None:
        """Set K, R to matrix and T to vector.

        Args:
            K (Union[list, np.ndarray, None]):
                Nested list of float32, 4x4 or 3x3 K mat.
                Defaults to None, intrisic will not be changed.
            R (Union[list, np.ndarray, None]):
                Nested list of float32, 3x3 R mat.
                Defaults to None, extrisic_r will not be changed.
            T (Union[list, np.ndarray, None]):
                List of float32, T vector.
                Defaults to None, extrisic_t will not be changed.
            world2cam (Union[bool, None], optional):
                Whether the R, T transform points from world space
                to camera space.
                Defaults to None, self.world2cam will not be changed.
        """
        if K is not None:
            if len(K) == 4:
                self.intrinsic = np.asarray(K)
            else:
                self.set_intrinsic(mat3x3=K, perspective=True)
        if R is not None:
            self.extrinsic_r = np.asarray(R)
        if T is not None:
            self.extrinsic_t = np.asarray(T)
        if world2cam is not None:
            assert isinstance(world2cam, bool)
            self.world2cam = world2cam

    def inverse_extrinsic(self) -> None:
        """Inverse the direction of extrinsics, between world to camera and
        camera to world."""
        r_mat = np.asarray(self.extrinsic_r)
        t_vec = np.asarray(self.extrinsic_t)
        r_mat = np.linalg.inv(r_mat).reshape(3, 3)
        t_vec = -np.dot(r_mat, t_vec).reshape(3)
        self.extrinsic_r = r_mat
        self.extrinsic_t = t_vec
        self.world2cam = not self.world2cam

    def intrinsic33(self) -> np.ndarray:
        """Get an intrinsic matrix in shape (3, 3).

        Returns:
            ndarray: An ndarray of intrinsic matrix.
        """
        return super().intrinsic33()

    def get_intrinsic(self, k_dim: int = 3) -> list:
        """Get intrinsic K matrix.

        Args:
            k_dim (int, optional):
                If 3, returns a 3x3 mat.
                Else if 4, returns a 4x4 mat.
                Defaults to 3.

        Raises:
            ValueError: k_dim is neither 3 nor 4.

        Returns:
            list: Nested list of float32, 4x4 or 3x3 K mat.
        """
        if k_dim == 4:
            return self.intrinsic.tolist()
        elif k_dim == 3:
            return super().intrinsic33().tolist()
        else:
            self.logger.error('k_dim is neither 3 nor 4.')
            raise ValueError

    def get_extrinsic_r(self) -> list:
        """Get extrinsic rotation matrix.

        Returns:
            list: Nested list of float32, 3x3 R mat.
        """
        return self.extrinsic_r.tolist()

    def get_extrinsic_t(self) -> list:
        """Get extrinsic translation vector.

        Returns:
            list: Nested list of float32, T vec of length 3.
        """
        return self.extrinsic_t.reshape(3).tolist()

    def SaveFile(self, filename: str) -> int:
        """Dump camera name and parameters to a json file.

        Args:
            filename (str):
                Path to the dumped json file.

        Returns:
            int: returns 0.
        """
        return BaseCameraParameter_cpp.SaveFile(self, filename)

    def dump(self, filename: str) -> None:
        """Dump camera name and parameters to a json file.

        Args:
            filename (str):
                Path to the dumped json file.

        Raises:
            RuntimeError: Fail to dump a json file.
        """
        if not self.SaveFile(filename):
            self.logger.error('Fail to dump a json file.')
            raise RuntimeError

    def LoadFile(self, filename: str) -> bool:
        """Load camera name and parameters from a dumped json file.

        Args:
            filename (str):
                Path to the dumped json file.

        Returns:
            bool: True if load succeed.
        """
        return BaseCameraParameter_cpp.LoadFile(self, filename)

    def load(self, filename: str) -> None:
        """Load camera name and parameters from a dumped json file.

        Args:
            filename (str):
                Path to the dumped json file.

        Raises:
            FileNotFoundError: File not found at filename.
            ValueError: Content in filename is not correct.
        """
        if not os.path.exists(filename):
            self.logger.error(f'File not found at {filename}.')
            raise FileNotFoundError
        test_cam = self.__class__()
        load_test = test_cam.LoadFile(filename)
        if load_test:
            self.LoadFile(filename)
        else:
            self.logger.error('File content is not correct.')
            raise ValueError

    def clone(self) -> 'BaseCameraParameter':
        """Clone a new CameraPrameter instance like self.

        Returns:
            BaseCameraParameter
        """
        new_cam_param = self.__class__(
            K=copy.deepcopy(self.get_intrinsic(k_dim=4)),
            R=copy.deepcopy(self.extrinsic_r),
            T=copy.deepcopy(self.extrinsic_t),
            name=self.name,
            height=self.height,
            width=self.width,
            world2cam=self.world2cam,
            convention=self.convention,
            logger=self.logger)
        return new_cam_param

    @classmethod
    def fromfile(cls, filename: str) -> 'BaseCameraParameter':
        """Construct a camera parameter data structure from a json file.

        Args:
            filename (str):
                Path to the dumped json file.

        Returns:
            CameraParameter:
                An instance of CameraParameter class.
        """
        ret_cam = cls()
        ret_cam.load(filename)
        return ret_cam



import json
import logging
from typing import List, Union

from xrprimer.data_structure.camera import PinholeCameraParameter
from xrprimer_cpp import \
    VectorPinholeCameraParameter as VectorPinholeCameraParameter_cpp
from xrprimer_cpp import calibrator as calibrator_cpp
from .base_calibrator import BaseCalibrator


class MviewPinholeCalibrator(BaseCalibrator):

    def __init__(self,
                 chessboard_width: int,
                 chessboard_height: int,
                 chessboard_square_size: int,
                 calibrate_intrinsic: bool = False,
                 calibrate_extrinsic: bool = True,
                 logger: Union[None, str, logging.Logger] = None) -> None:
        """Mview extrinsic calibrator for distorted fisheye cameras.

        Args:
            chessboard_width (int):
                How many internal corners along the
                horizontal edge of the chessboard.
            chessboard_height (int):
                How many internal corners along the
                vertical edge of the chessboard.
            chessboard_square_size (int):
                The edge length of a unit square in millimeter.
            calibrate_intrinsic (bool, optional):
                Whether to calibrate intrinsic. Defaults to False.
            calibrate_extrinsic (bool, optional):
                Whether to calibrate extrinsic. Defaults to True.
            logger (Union[None, str, logging.Logger], optional):
                Logger for logging. If None, root logger will be selected.
                Defaults to None.
        """
        BaseCalibrator.__init__(self, logger=logger)
        self.chessboard_width = chessboard_width
        self.chessboard_height = chessboard_height
        self.chessboard_square_size = chessboard_square_size
        if calibrate_intrinsic:
            self.logger.error('Intrinsic calibration not implemented yet.')
            raise NotImplementedError
        if (calibrate_intrinsic or calibrate_extrinsic) is False:
            self.logger.error(
                'Please select at least one parameter to calibrate.')
            raise ValueError
        self.calibrate_intrinsic = calibrate_intrinsic
        self.calibrate_extrinsic = calibrate_extrinsic

    def calibrate(
        self,
        frames: List[List[str]],
        pinhole_param_list: List[PinholeCameraParameter],
    ) -> List[PinholeCameraParameter]:
        """Calibrate multi-PinholeCameraParameters with a chessboard.

        Args:
            frames (List[List[str]]):
                A nested list of image paths. The shape is
                [n_frame, n_view], and each element is the path to
                an image file. '' stands for an empty image.
            pinhole_param_list (List[PinholeCameraParameter]):
                A list of PinholeCameraParameters. Intrinsic matrix
                is necessary for calibration.

        Returns:
            List[PinholeCameraParameter]:
                A list of calibrated pinhole cameras, name, logger,
                resolution will be kept.
        """

        if len(frames) <= 0:
            self.logger.error('Frames are necessary for pinhole extrinsic' +
                              ' calibration.')
            raise ValueError
        if len(frames[0]) != len(pinhole_param_list):
            self.logger.error(
                'n_view of frames must be equal to len(pinhole_param_list).')
            raise ValueError
        chessboard_config_dict = dict(
            chessboard_width=self.chessboard_width,
            chessboard_height=self.chessboard_height,
            chessboard_square_size=self.chessboard_square_size)
        ret_list = []
        if self.calibrate_extrinsic:
            chessboard_config_str = json.dumps(chessboard_config_dict)
            pinhole_vector = VectorPinholeCameraParameter_cpp(
                pinhole_param_list)
            calibrator_cpp.CalibrateMultiPinholeCamera(chessboard_config_str,
                                                       frames, pinhole_vector)
            for cam_idx, pinhole_param_cpp in enumerate(pinhole_vector):
                pinhole_param = PinholeCameraParameter(
                    name=pinhole_param_list[cam_idx].name,
                    K=pinhole_param_cpp.intrinsic,
                    R=pinhole_param_cpp.extrinsic_r,
                    T=pinhole_param_cpp.extrinsic_t,
                    height=pinhole_param_list[cam_idx].height,
                    width=pinhole_param_list[cam_idx].width,
                    world2cam=False,
                    convention='opencv',
                    logger=pinhole_param_list[cam_idx].logger)
                ret_list.append(pinhole_param)
        return ret_list



from .version import __version__


def digit_version(version_str):
    digit_version = []
    for x in version_str.split('.'):
        if x.isdigit():
            digit_version.append(int(x))
        elif x.find('rc') != -1:
            patch_version = x.split('rc')
            digit_version.append(int(patch_version[0]) - 1)
            digit_version.append(int(patch_version[1]))
    return digit_version


__all__ = ['__version__']



import numpy as np


def bgr2rgb(input_array: np.ndarray, color_dim: int = -1) -> np.ndarray:
    """Convert image array of any shape between BGR and RGB.

    Args:
        input_array (np.ndarray):
            An array of images. The shape could be:
            [h, w, n_ch], [n_frame, h, w, n_ch],
            [n_view, n_frame, h, w, n_ch], etc.
        color_dim (int, optional):
            Which dim is the color channel. Defaults to -1.

    Returns:
        np.ndarray
    """
    r_slice_list = [
        slice(None),
    ] * len(input_array.shape)
    b_slice_list = [
        slice(None),
    ] * len(input_array.shape)
    r_slice_list[color_dim] = slice(0, 1, 1)
    b_slice_list[color_dim] = slice(2, 3, 1)
    b_backup = input_array[tuple(b_slice_list)].copy()
    input_array[tuple(b_slice_list)] = input_array[tuple(r_slice_list)]
    input_array[tuple(r_slice_list)] = b_backup
    return input_array



import os
import re
import shutil
import subprocess
import sys

from setuptools import Extension, find_packages, setup
from setuptools.command.build_ext import build_ext

# Our python package root dir is python/
PACKAGE_DIR = 'python'

PACKAGES = find_packages(PACKAGE_DIR)

# Convert distutils Windows platform specifiers to CMake -A arguments
PLAT_TO_CMAKE = {
    'win32': 'Win32',
    'win-amd64': 'x64',
    'win-arm32': 'ARM',
    'win-arm64': 'ARM64',
}


def parse_version_from_file(filepath):
    """Parse version txt into string and tuple.

    Args:
        filepath (str): The version filepath.
    Returns:
        str: The version string, e.g., "0.4.0"
        dict: The version info, e.g.,
            {
                'XRPRIMER_VERSION_MAJOR': 0,
                'XRPRIMER_VERSION_MINOR': 4,
                'XRPRIMER_VERSION_PATCH': 0,
            }
    """
    keywords = [
        'XRPRIMER_VERSION_MAJOR', 'XRPRIMER_VERSION_MINOR',
        'XRPRIMER_VERSION_PATCH'
    ]
    version_info = {}
    version_list = []
    with open(filepath) as f:
        content = f.read()
        for keyword in keywords:
            regex = rf'{keyword}\s*([0-9]+)'
            obj = re.search(regex, content)
            vid = obj.group(1)
            version_list.append(vid)
            assert keyword not in version_info
            version_info[keyword] = int(vid)
    version = '.'.join(version_list)
    return version, version_info


VERSION, VERSION_INFO = parse_version_from_file('version.txt')


# A CMakeExtension needs a sourcedir instead of a file list.
# The name must be the _single_ output extension from the CMake build.
# If you need multiple extensions, see scikit-build.
class CMakeExtension(Extension):

    def __init__(self, name, sourcedir=''):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


def clean_cmake(folders):
    for folder in folders:
        if not os.path.exists(folder):
            continue
        shutil.rmtree(folder)


class CMakeBuild(build_ext):

    def build_extension(self, ext):
        extdir = os.path.abspath(
            os.path.dirname(self.get_ext_fullpath(ext.name)))

        # required for auto-detection & inclusion of auxiliary "native" libs
        if not extdir.endswith(os.path.sep):
            extdir += os.path.sep

        debug = int(os.environ.get('DEBUG',
                                   0)) if self.debug is None else self.debug
        cfg = 'Debug' if debug else 'Release'

        # CMake lets you override the generator - we need to check this.
        # Can be set with Conda-Build, for example.
        cmake_generator = os.environ.get('CMAKE_GENERATOR', '')

        # Set Python_EXECUTABLE instead if you use PYBIND11_FINDPYTHON
        # EXAMPLE_VERSION_INFO shows you how to pass a value into the C++ code
        # from Python.
        cmake_args = [
            f'-DCMAKE_LIBRARY_OUTPUT_DIRECTORY={extdir}',
            f'-DPYTHON_EXECUTABLE={sys.executable}',
            f'-DCMAKE_BUILD_TYPE={cfg}',  # not used on MSVC, but no harm
        ]
        build_args = []
        # Adding CMake arguments set as environment variable
        # (needed e.g. to build for ARM OSx on conda-forge)
        if 'CMAKE_ARGS' in os.environ:
            cmake_args += [
                item for item in os.environ['CMAKE_ARGS'].split(' ') if item
            ]

        # Enable test
        cmake_args += ['-DENABLE_TEST=OFF']
        # Build external from pre-built libs by default
        ret = os.system('conan remote list | grep xrlab')
        if ret == 0:  # remote exists
            cmake_args += ['-DBUILD_EXTERNAL=OFF']
        else:
            cmake_args += ['-DBUILD_EXTERNAL=ON']

        # Pass version
        for key, version in VERSION_INFO.items():
            assert key.isupper(), f'{key} is expected to be uppercase'
            cmake_args += [f'-D{key}={version}']

        if self.compiler.compiler_type != 'msvc':
            # Using Ninja-build since it a) is available as a wheel and b)
            # multithreads automatically. MSVC would require all variables be
            # exported for Ninja to pick it up, which is a little tricky to do.
            # Users can override the generator with CMAKE_GENERATOR in CMake
            # 3.15+.
            if not cmake_generator or cmake_generator == 'Ninja':
                try:
                    import ninja  # noqa: F401

                    ninja_executable_path = os.path.join(
                        ninja.BIN_DIR, 'ninja')
                    cmake_args += [
                        '-GNinja',
                        f'-DCMAKE_MAKE_PROGRAM:FILEPATH={ninja_executable_path}',  # noqa: E501
                    ]
                except ImportError:
                    pass

        else:

            # Single config generators are handled "normally"
            single_config = any(x in cmake_generator
                                for x in {'NMake', 'Ninja'})

            # CMake allows an arch-in-generator style for backward
            # compatibility.
            contains_arch = any(x in cmake_generator for x in {'ARM', 'Win64'})

            # Specify the arch if using MSVC generator, but only if it doesn't
            # contain a backward-compatibility arch spec already in the
            # generator name.
            if not single_config and not contains_arch:
                cmake_args += ['-A', PLAT_TO_CMAKE[self.plat_name]]

            # Multi-config generators have a different way to specify configs
            if not single_config:
                cmake_args += [
                    f'-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{cfg.upper()}={extdir}'
                ]
                build_args += ['--config', cfg]

        if sys.platform.startswith('darwin'):
            # Cross-compile support for macOS - respect ARCHFLAGS if set
            archs = re.findall(r'-arch (\S+)', os.environ.get('ARCHFLAGS', ''))
            if archs:
                cmake_args += [
                    '-DCMAKE_OSX_ARCHITECTURES={}'.format(';'.join(archs))
                ]

        # Set CMAKE_BUILD_PARALLEL_LEVEL to control the parallel build level
        # across all generators.
        if 'CMAKE_BUILD_PARALLEL_LEVEL' not in os.environ:
            # self.parallel is a Python 3 only way to set parallel jobs by hand
            # using -j in the build_ext call, not supported by pip or
            # PyPA-build.
            if not hasattr(self, 'parallel') or not self.parallel:
                # Manually use 1/4 total cpu for compilation
                import multiprocessing as mp
                cpu_count = mp.cpu_count()
                self.parallel = max(1, int(cpu_count / 4))
                print(f'{self.parallel} cpu cores are used for compilation')
            if hasattr(self, 'parallel') and self.parallel:
                # CMake 3.12+ only.
                build_args += [f'-j{self.parallel}']

        install_dir = 'install'
        build_temp = self.build_temp
        clean_cmake(folders=['_deps', '_exts', build_temp, install_dir])
        subprocess.check_call(['cmake', '-S.', '-B', build_temp] + cmake_args)
        subprocess.check_call(
            ['cmake', '--build', build_temp, '--target', install_dir] +
            build_args)


def readme():
    with open('./README.md', encoding='utf-8') as f:
        content = f.read()
    return content


def parse_requirements(fname='requirements.txt', with_version=True):
    """Parse the package dependencies listed in a requirements file but strips
    specific versioning information.

    Args:
        fname (str): path to requirements file
        with_version (bool, default=False): if True include version specs

    Returns:
        List[str]: list of requirements items

    CommandLine:
        python -c "import setup; print(setup.parse_requirements())"
    """
    require_fpath = fname

    def parse_line(line):
        """Parse information from a line in a requirements text file."""
        if line.startswith('-r '):
            # Allow specifying requirements in other files
            target = line.split(' ')[1]
            for info in parse_require_file(target):
                yield info
        else:
            info = {'line': line}
            if line.startswith('-e '):
                info['package'] = line.split('#egg=')[1]
            else:
                # Remove versioning from the package
                pat = '(' + '|'.join(['>=', '==', '>']) + ')'
                parts = re.split(pat, line, maxsplit=1)
                parts = [p.strip() for p in parts]

                info['package'] = parts[0]
                if len(parts) > 1:
                    op, rest = parts[1:]
                    if ';' in rest:
                        # Handle platform specific dependencies
                        # http://setuptools.readthedocs.io/en/latest/setuptools.html#declaring-platform-specific-dependencies
                        version, platform_deps = map(str.strip,
                                                     rest.split(';'))
                        info['platform_deps'] = platform_deps
                    else:
                        version = rest  # NOQA
                    info['version'] = (op, version)
            yield info

    def parse_require_file(fpath):
        with open(fpath, 'r') as f:
            for line in f.readlines():
                line = line.strip()
                if line and not line.startswith('#'):
                    for info in parse_line(line):
                        yield info

    def gen_packages_items():
        if os.path.exists(require_fpath):
            for info in parse_require_file(require_fpath):
                parts = [info['package']]
                if with_version and 'version' in info:
                    parts.extend(info['version'])
                if not sys.version.startswith('3.4'):
                    # apparently package_deps are broken in 3.4
                    platform_deps = info.get('platform_deps')
                    if platform_deps is not None:
                        parts.append(';' + platform_deps)
                item = ''.join(parts)
                yield item

    packages = list(gen_packages_items())
    return packages


setup(
    name='xrprimer',
    version=VERSION,
    description='description',
    long_description=readme(),
    long_description_content_type='text/markdown',
    author='OpenXRLab',
    author_email='openxrlab@sensetime.com',
    keywords='xrprimer',
    url='https://gitlab.bj.sensetime.com/openxrlab/xrprimer',
    package_dir={'': PACKAGE_DIR},
    packages=PACKAGES,
    include_package_data=True,
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    license='Apache License 2.0',
    tests_require=parse_requirements('requirements/test.txt'),
    install_requires=parse_requirements('requirements/runtime.txt'),
    ext_modules=[CMakeExtension('xrprimer_cpp')],
    cmdclass={'build_ext': CMakeBuild},
    zip_safe=False)



import glob
import os
import shutil

import cv2
import mmcv
import numpy as np
import pytest

from xrprimer.data_structure.camera import PinholeCameraParameter
from xrprimer.ops.projection.builder import build_projector

input_dir = 'test/data/ops/test_projection'
output_dir = 'test/data/output/ops/test_projection'


@pytest.fixture(scope='module', autouse=True)
def fixture():
    if os.path.exists(output_dir):
        shutil.rmtree(output_dir)
    os.makedirs(output_dir, exist_ok=False)


def test_opencv_projector():
    keypoints3d = np.load(
        os.path.join(input_dir, 'keypoints3d.npz'), allow_pickle=True)
    keypoints3d = dict(keypoints3d)
    n_view = len(glob.glob(os.path.join(input_dir, '*.json')))
    cam_param_list = []
    for cam_idx in range(n_view):
        cam_param_path = os.path.join(input_dir, f'cam_{cam_idx:03d}.json')
        cam_param = PinholeCameraParameter()
        cam_param.load(cam_param_path)
        cam_param_list.append(cam_param)
    projector_config = dict(
        mmcv.Config.fromfile('config/ops/projection/opencv_projector.py'))
    projector_config['camera_parameters'] = cam_param_list
    projector = build_projector(projector_config)
    point = np.array(keypoints3d['keypoints'][0, 0, 0, :3])
    # test project numpy
    projected_points = projector.project_single_point(point)
    assert projected_points.shape == (n_view, 2)
    # test project list
    projected_points = projector.project_single_point(point.tolist())
    assert projected_points.shape == (n_view, 2)
    # test project tuple
    projected_points = projector.project_single_point((0, 1, 2))
    assert projected_points.shape == (n_view, 2)
    points3d = keypoints3d['keypoints'][0, 0, :, :3]
    points3d_mask = np.expand_dims(keypoints3d['mask'][0, 0, :], axis=-1)
    n_point = points3d.shape[0]
    # test project numpy
    assert isinstance(points3d, np.ndarray)
    projected_points = projector.project(points3d)
    assert projected_points.shape == (n_view, n_point, 2)
    # test project list
    projected_points = projector.project(points3d.tolist())
    assert projected_points.shape == (n_view, n_point, 2)
    projected_points = projector.project(points3d, points_mask=points3d_mask)
    for cam_idx in range(n_view):
        canvas = cv2.imread(os.path.join(input_dir, f'{cam_idx:06}.png'))
        valid_idxs = np.where(points3d_mask == 1)
        for point_idx in valid_idxs[0]:
            cv2.circle(
                img=canvas,
                center=projected_points[cam_idx, point_idx].astype(np.int32),
                radius=2,
                color=(0, 0, 255))
        cv2.imwrite(
            filename=os.path.join(output_dir, f'projected_kps_{cam_idx}.jpg'),
            img=canvas)
